from __future__ import annotations

import itertools
import operator
from collections.abc import (
    Generator,
    Sequence,
    Mapping,
    Container,
    Callable,
    Iterable,
    Hashable,
)
from numbers import Real
from typing import Optional, Any


Vector = Sequence[int]  # api.rst: documented manually
Vectors = Sequence[Vector]
Limits = Sequence[tuple[int, int]]  # api.rst: documented manually


class Position(tuple[int]):
    """A position in an n-dimensional array. It is initialized
    by a sequence of integer coordinates, e.g., a list or tuple."""

    @classmethod
    def at(cls, *coordinates: int):
        """Factory method that creates a position from integer *coordinates*
        given as separate parameters."""
        return Position(coordinates)

    def __add__(self, other: Vector) -> [Position]:
        """Add the *other* vector to the position"""
        return Position(map(sum, zip(self, other)))

    def __sub__(self, other: Vector) -> [Position]:
        """Subtract the *other* vector from the position"""
        return Position(map(operator.sub, self, other))

    def manhattan_distance(self, other: Vector) -> int:
        """Manhattan distance of the *other* vector from the
        given position vector"""
        return sum(abs(coordinate_diff) for coordinate_diff in self - other)

    def is_in_cuboid(self, limits: Limits) -> bool:
        """Returns true if each coordinate of the position is inside
        the respective range defined by the *limits* sequence.
        """
        return all(
            low_limit <= coordinate < high_limit
            for coordinate, (low_limit, high_limit) in zip(self, limits)
        )

    def wrap_to_cuboid(self, limits: Limits) -> Position:
        """If a coordinate of the position is outside its respective
        limit range *(from, to)* of the *limits* sequence, add or
        subtract the size (to - from + 1) of the limit range as often as
        necessary to correct this.
        """
        coordinates = []
        for coordinate, (low_limit, high_limit) in zip(self, limits):
            if coordinate < low_limit:
                window_size = high_limit - low_limit
                coordinate = high_limit - (high_limit - coordinate) % window_size
            if coordinate >= high_limit:
                window_size = high_limit - low_limit
                coordinate = (coordinate - low_limit) % window_size + low_limit
            coordinates.append(coordinate)
        return Position(coordinates)

    @staticmethod
    def moves(
        dimensions: int = 2,
        diagonals: bool = False,
        zero_move: bool = False,
    ) -> Vectors:
        """
        Generate vectors of moves to neighbor coordinates in a n-dimensional
        Array in sorted order, i.e., vectors that differ in each coordinate
        by -1, 0 or 1. Default: Leave out the "zero move" *(0, ...)* and
        the "diagonal" moves (no zeros in any coordinate)."""
        moves = []
        for move in itertools.product(range(-1, 2), repeat=dimensions):
            if not zero_move and all(coordinate == 0 for coordinate in move):
                continue
            if not diagonals and all(coordinate != 0 for coordinate in move):
                continue
            moves.append(move)
        return moves

    def neighbors(
        self,
        moves: Iterable[Vector],
        limits: Optional[Limits] = None,
        wrap: bool = False,
    ) -> Generator[Position]:
        # noinspection PyShadowingNames
        """
        Generate coordinates of direct neighbors of a given position in a n-dimensional matrix.
        If pairs of lower and upper limits per coordinate are given, the generated coordinates are
        filtered or wrapped (if wrap == True) to stay within the limits. Diagonal neighbors are
        omitted on request. The neighbors are generated in sorted order.
        """
        # inspection PyShadowingNames
        for move in moves:
            neighbor_position = self + move
            if limits is None:
                if wrap is True:
                    raise RuntimeError("Limits for Option wrap missing")
                yield neighbor_position
            elif wrap:
                yield neighbor_position.wrap_to_cuboid(limits)
            else:
                if neighbor_position.is_in_cuboid(limits):
                    yield neighbor_position


class Array:
    def __init__(self, nested_sequences, dimensions: int = 2):
        """An n-dimensional array.

        Based on *nested sequences* that, up to up to a given number of
        *dimensions*, define its content. Data in nestings deeper than
        that is interpreted as part of the content of an array cell.

        Coordinates of a position in the array are meant in the order from "outer"
        to "inner" sequences along the nesting. Coordinates can be given as
        Vector, especially as Position.
        """
        self.content = nested_sequences
        self.dimensions = dimensions

    def size(self) -> Sequence[int]:
        # noinspection PyShadowingNames
        """Calculates the size of the array per dimension."""
        # inspection PyShadowingNames
        size = []
        area = self.content
        for d in range(self.dimensions):
            size.append(len(area))
            area = area[0]
        return size

    def limits(self) -> Limits:
        """Calculates coordinate limits per dimension as tuples
        (from, to). Since the array consists of nested sequences,
        *from* is always zero and *to* equals the size of the array
        in this dimension."""
        return [(0, upper) for upper in self.size()]

    def mutable_copy(self) -> Array:
        """Creates a mutable copy of the array."""

        def _writable(area, dimensions):
            if dimensions > 1:
                return [_writable(sub_area, dimensions - 1) for sub_area in area]
            else:
                return list(area)

        return Array(_writable(self.content, self.dimensions), self.dimensions)

    def __getitem__(self, position: Vector):
        """Gets the content at the given position. This allows for using
        the expression syntax *array[...]*."""
        i = self.content
        for coordinate in position:
            i = i[coordinate]
        return i

    def __setitem__(self, position: Vector, content: Any):
        # noinspection PyShadowingNames
        """Sets the content at the given position. Assumption: The nested
        sequence the array is build upon is mutable in its last dimension
        (i.e., a list). Allows for using the assignment syntax
        *array[...] = ...* ."""
        # inspection PyShadowingNames
        field = self.content
        for coordinate in position[:-1]:
            field = field[coordinate]
        field[position[-1]] = content

    def items(self) -> Generator[tuple[Position, Any]]:
        """Iterates positions and content."""
        # Attention: Method signature is manually documented, update also there

        def _items_in_dimension(area, dimensions):
            if dimensions == 1:
                for coordinate, sub_area in enumerate(area):
                    yield (coordinate,), sub_area
            else:
                for coordinate, sub_area in enumerate(area):
                    for sub_vector, content in _items_in_dimension(
                        sub_area, dimensions - 1
                    ):
                        yield (coordinate,) + sub_vector, content

        return (
            (Position(coordinates), content)
            for coordinates, content in _items_in_dimension(
                self.content, self.dimensions
            )
        )

    def findall(self, content: Container[Any]) -> tuple[Position]:
        """Finds content in array and return found positions.
        The content is given in some container, i.e., a set.
        """
        # Attention: Method signature is manually documented, update also there

        content_set = set(content)

        def find_in_dimension(p_matrix, p_dimensions) -> Generator[tuple[int]]:
            if p_dimensions == 1:
                for coordinate, cell_content in enumerate(p_matrix):
                    if cell_content in content_set:
                        yield (coordinate,)
            else:
                for coordinate, sub_area in enumerate(p_matrix):
                    for found in find_in_dimension(sub_area, p_dimensions - 1):
                        yield (coordinate,) + found

        return tuple(
            Position(v) for v in find_in_dimension(self.content, self.dimensions)
        )

    def next_vertices_from_forbidden(
        self,
        forbidden: [Iterable[Hashable]],
        wrap: bool = False,
        diagonals: bool = False,
    ) -> [Callable]:
        # noinspection PyShadowingNames
        """Returns a NextVertices function for traversal strategies, based on
        your choice of when positions qualifies as neighbors (goals of a
        move) of a given position (options *wrap* and *diagonals*) and whether
        such a move is allowed w.r.t. the content of the array at this
        position (parameter *forbidden*).
        """
        # inspection PyShadowingNames
        forbidden_content = set(forbidden)
        limits = self.limits()
        moves = Position.moves(dimensions=self.dimensions, diagonals=diagonals)

        def next_vertices(position, _):
            for neighbor in position.neighbors(moves=moves, limits=limits, wrap=wrap):
                if self[neighbor] not in forbidden_content:
                    yield neighbor

        return next_vertices

    def next_edges_from_cell_weights(
        self,
        content_to_weight: [Mapping[Any, Real]],
        wrap: bool = False,
        diagonals: bool = False,
    ) -> [Callable]:
        # noinspection PyShadowingNames
        """Returns a NextEdges function for traversal strategies, based on
        your choice of when positions qualifies as neighbors (goals of a
        move) of a given position (options *wrap* and *diagonals*) and what
        weight such a move has w.r.t. the content of the matrix at this
        position (parameter *content_to_weight*, and no assigned weight means
        the move is impossible).
        """
        # inspection PyShadowingNames
        limits = self.limits()
        moves = Position.moves(dimensions=self.dimensions, diagonals=diagonals)

        def next_edges(vector, _):
            for neighbor in vector.neighbors(moves=moves, limits=limits, wrap=wrap):
                weight = content_to_weight.get(self[neighbor], None)
                if weight is not None:
                    yield neighbor, weight

        return next_edges



from nographs.types import *
from nographs.paths import *
from nographs.strategies import *
from nographs.matrix_gadgets import *
from nographs.edge_gadgets import *

__all__ = (
    "Vertex",
    "VertexIterator",
    "NextVertices",
    "NextEdges",
    "VertexToID",
    "Paths",
    "PathsOfUnlabeledEdges",
    "PathsOfLabeledEdges",
    "Traversal",
    "TraversalBreadthFirst",
    "TraversalDepthFirst",
    "TraversalTopologicalSort",
    "TraversalShortestPaths",
    "TraversalAStar",
    "TraversalMinimumSpanningTree",
    "adapt_edge_index",
    "adapt_edge_iterable",
    "Vector",
    "Limits",
    "Position",
    "Array",
)



from __future__ import annotations

import collections
import itertools
from collections.abc import Sequence, Mapping, Callable, Iterable
from typing import Union


# ---------- functions for handling test data given as some kind of edges -------------


def adapt_edge_index(
    index: Union[Mapping, Sequence], *, add_inverted: bool, labeled: bool
) -> Callable:
    """
    Reads a test graph from a Mapping (e.g. a Dict) or from a Sequence (e.g. a tuple or list,
    if you use integers as your vertices) and provide a neighbor function from that data.
    Typically only used for test purposes.

    :param index: Mapping or Sequence with vertices as key resp. index. For each vertex, you provide
        an Iterable of adjacent vertices (this defines unlabeled edges) or an Iterable of the
        labeled edges that start at this vertex. A labeled edge to some neighbor need to be a
        Sequence of the form (neighbor, additional data...).

    :param add_inverted: The resulting neighbor function will yield the edges of m.
        Additionally, if add_inverted_edges is True, it will provide the same edges inverted (start
        end end vertex exchanged).

    :param labeled: Give True, if *index* provides labeled edges. In this case, you can use the
        resulting neighbor function as next_edges function. Give False, if m provides neighbor
        vertices for a given vertex. In this case, you can use the resulting neighbor function as
        next_vertices function.

    :return: Neighbor function that can be used as parameter for one of the traversal algorithms.
    """

    if add_inverted:
        if isinstance(index, Mapping):
            return adapt_edge_iterable(
                (
                    (v1,) + e_to if labeled else (v1, e_to)
                    for v1, edges_to in index.items()
                    for e_to in edges_to
                ),
                add_inverted=add_inverted,
                labeled=labeled,
            )
        if isinstance(index, Sequence):
            return adapt_edge_iterable(
                (
                    (v1,) + e_to if labeled else (v1, e_to)
                    for edges_to, v1 in zip(index, itertools.count())
                    for e_to in edges_to
                ),
                add_inverted=add_inverted,
                labeled=labeled,
            )
        raise ValueError("graph must be Mapping or Sequence")

    def get(vertex, _):
        try:
            return index[vertex]
        except KeyError:
            return ()
        except IndexError:
            return ()

    return get


def adapt_edge_iterable(
    edges: Iterable[Sequence], *, add_inverted: bool, labeled: bool
) -> Callable:
    """
    Reads a graph from an Iterable of edges and provide a neighbor function from that data.
    Typically used only for test purposes.

    :param edges: The edges of your graph, each as Sequence (start_vertex, end_vertex,
        optional_more_data...)

    :param add_inverted: The resulting neighbor function will yield your edges.
        Additionally, if add_inverted_edges is True, it will provide the same edges inverted (start
        and end vertices exchanged). Note: when this option is used, a copy of your graph will be
        held by the returned neighbor function.

    :param labeled: If set to True, the resulting neighbor function will yield labeled edges (
        to_vertex, additional data ...) for a given vertex and can be used as next_edges function.
        If set to False, it will yield neighbor vertices, can be used as next_vertices function,
        and additional data that your edges may contain will be ignored.

    :return: Neighbor function that can be used as parameter for one of the traversal algorithms.
    """
    edge_dict = collections.defaultdict(list)
    if add_inverted:
        if labeled:
            # Labeled edges are provided and all data should be used
            for from_vertex, to_vertex, *data in edges:
                edge_dict[from_vertex].append((to_vertex, *data))
                edge_dict[to_vertex].append((from_vertex, *data))
        else:
            # Only start and end vertices of the edge should be used. Optional additional data is
            # ignored.
            for from_vertex, to_vertex, *data in edges:
                edge_dict[from_vertex].append(to_vertex)
                edge_dict[to_vertex].append(from_vertex)
    else:
        for edge in edges:
            if labeled:
                from_vertex, rest = edge[0], edge[1:]
                edge_dict[from_vertex].append(rest)
            else:
                from_vertex, to_vertex = edge[0], edge[1]
                edge_dict[from_vertex].append(to_vertex)
    return adapt_edge_index(edge_dict, add_inverted=False, labeled=labeled)



if __name__ == "__main__":
    import doctest
    import unittest
    import coverage
    import importlib
    import pathlib

    # Start recording coverage
    cov = coverage.Coverage()
    cov.start()

    # Create empty TestSuite
    test_suite = unittest.TestSuite()

    # Unittests from doc tests in modules of package
    modules = ("types", "strategies", "matrix_gadgets", "edge_gadgets", "paths")
    for f in modules:
        temp_module = importlib.import_module("." + f, "nographs")
        test_suite.addTests(doctest.DocTestSuite(temp_module))

    # Unittests from doc tests in module with special test cases
    __import__("test_code")
    test_suite.addTests(doctest.DocTestSuite("test_code"))

    # Unittests from doc tests
    p = pathlib.Path('docs/source')
    for file_path in p.glob('*.rst'):
        test_suite.addTests(doctest.DocFileSuite(str(file_path.resolve())))

    unittest.TextTestRunner().run(test_suite)

    # Stop recording coverage, create HTML from results
    cov.stop()
    cov.save()
    cov.html_report()



from __future__ import annotations

import collections
import itertools
import copy
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterator, Iterable, Hashable
from heapq import heapify, heappop, heappush
from numbers import Real
from typing import TypeVar, Optional

from nographs import Vertex, VertexIterator, NextVertices, NextEdges, VertexToID
from nographs import Paths, PathsOfUnlabeledEdges, PathsOfLabeledEdges


# --------------- internal support functions -------------


def _iter_start_ids(
    start_vertices: Iterator[Vertex],
    vertex_to_id: Optional[Callable[[Vertex], Hashable]],
) -> Iterable[Hashable]:
    if vertex_to_id:
        return (vertex_to_id(vertex) for vertex in start_vertices)

    return start_vertices


def _define_visited(
    already_visited: Optional[set[Hashable]],
    iter_start_ids: Iterable[Hashable],
    is_tree: bool,
) -> set[Hashable]:
    """Use already_visited, if provided, for storing visited vertices, and otherwise new dict.
    Mark start vertices as visited.
    """
    if already_visited is None:
        return set() if is_tree else set(iter_start_ids)

    if not is_tree:
        already_visited.update(iter_start_ids)
    return already_visited


def _create_paths(
    build_paths: bool,
    labeled_path: bool,
    labeled_edges: bool,
    vertex_to_id: Optional[VertexToID],
) -> tuple[Optional[Paths], Optional[dict], Optional[dict]]:
    """Translate from configuration of path generation to a function initializing the required
    thing"""

    if not build_paths:
        if labeled_path:
            raise RuntimeError("Option labeled_paths without option build_paths.")
        return None, None, None

    predecessors = dict()
    if not labeled_path:
        return (
            PathsOfUnlabeledEdges(predecessors, vertex_to_id),
            predecessors,
            None,
        )

    if not labeled_edges:
        raise RuntimeError("A labeled path can only be computed from labeled edges.")

    edge_data = dict()
    return (
        PathsOfLabeledEdges(predecessors, edge_data, vertex_to_id),
        predecessors,
        edge_data,
    )


class NoIterator:
    def __next__(self):
        raise RuntimeError("Traversal not started, iteration not possible")


# ----------- traversal strategies for unweighted graphs with or without edge labels ----------


class Traversal(ABC):
    """
    Abstract Class. Its subclasses provide methods to iterate through vertices and edges using some
    specific traversal strategies.
    """

    @abstractmethod
    def __init__(
        self,
        next_edge_or_vertex: Callable,
        labeled_edges: bool,
        is_tree: bool,
        vertex_to_id: Optional[VertexToID],
    ):
        # attributes of graph adaptation
        self._next_edge_or_vertex = next_edge_or_vertex
        self._labeled_edges = labeled_edges
        self._is_tree = is_tree
        self._vertex_to_id = vertex_to_id

        # general attributes set and needed by all traversal strategies
        self._generator = NoIterator()
        self._start_vertices = None
        self._build_paths = None
        self._labeled_paths = None
        self._calculation_limit = None

        # attributes for path data, needed by all traversal strategies
        self.paths = None
        self._predecessors = None
        self._edge_data = None

    def __iter__(
        self,
    ) -> VertexIterator:  # Do to a sphinx error, the enclosing [] necessary
        """
        Returns the iterator of a started traversal. This allows for using a `Traversal` in *for*
        loops or as parameter to a call of function *next()*.

        Subsequent calls return the same iterator again. This allows for using the same `Traversal`
        in subsequent *for* loops or *next()* calls, as long as the iterator is not exhausted.

        The iterator yields vertices reported by the traversal algorithm. When a vertex is
        reported, specific attributes of the traversal object contain additional data about the
        state of the traversal (see the API documentation of the respective subclass of
        `Traversal`)."""
        if not isinstance(type(self), type(Traversal)):
            raise RuntimeError("Method go can only be called on a Traversal object.")
        return self._generator

    def __next__(self) -> Vertex:
        """Returns the next vertex reported by the (started) traversal. Delegates to the iterator
        of the traversal."""
        return next(self._generator)

    def go_for_vertices_in(
        self, vertices: VertexIterator, fail_silently: bool = False
    ) -> VertexIterator:
        """
        For a started traversal, the method returns an iterator that fetches vertices
        from the traversal, reports a vertex if it is in *vertices*, and stops when all of the
        *vertices* have been found and reported. If the iterator has no more vertices to
        report (graph is exhausted) without having found all of the *vertices*, KeyError is
        raised, or the traversal just terminates, if a silent fail is demanded.

        Whenever a vertex is reported, specific attributes of the traversal object contain
        additional data about the state of the traversal (see the API documentation of the
        respective subclass of `Traversal`).
        """
        if not isinstance(type(self), type(Traversal)):
            raise RuntimeError(
                "Method go_for_vertices_in can only be called "
                + "on a Traversal object."
            )
        # In order to make the above check work, the following generator functionality
        # needs to be encapsulated in a local function

        def my_generator():
            vertex_set = (
                set(vertices)
                if self._vertex_to_id is None
                else set(self._vertex_to_id(vertex) for vertex in vertices)
            )

            v_count = len(vertex_set)
            for v in self._generator:
                if v not in vertex_set:
                    continue
                yield v
                v_count -= 1
                if v_count == 0:
                    break
            else:
                if not fail_silently:
                    raise KeyError("Not all of the given vertices have been found")

        return my_generator()

    def go_to(self, vertex: Vertex, fail_silently: bool = False) -> Optional[Vertex]:
        """
        For a started traversal, it walks through the graph, stops at *vertex* and returns it.
        If the traversal ends (traversal iterator is exhausted) without having found *vertex*,
        KeyError is raised, or None is returned, if fail_silently is True.

        When *vertex* is reported, specific attributes of the traversal object contain
        additional data about the state of the traversal (see the API documentation of the
        respective subclass of `Traversal`).
        """
        if not isinstance(type(self), type(Traversal)):
            raise RuntimeError("Method go_to can only be called on a Traversal object.")
        for v in self._generator:
            if v != vertex:
                continue
            return vertex
        else:
            if fail_silently:
                return None
            else:
                raise KeyError("Vertex not found, graph exhausted.")

    def _start_from(
        self,
        start_vertex: Optional[Vertex],
        start_vertices: Optional[Iterable[Vertex]],
        build_paths: bool,
        labeled_paths: bool,
        calculation_limit: Optional[int],
    ):
        if start_vertex is not None:
            if start_vertices is not None:
                raise RuntimeError("Both start_vertex and start_vertices provided.")
            self._start_vertices = (start_vertex,)
        else:
            if start_vertices is None:
                raise RuntimeError("Neither start_vertex and start_vertices provided.")
            self._start_vertices = start_vertices

        # Note: Detection of wrong option combinations for paths is implemented in _create_paths
        self._labeled_paths = labeled_paths
        self.paths, self._predecessors, self._edge_data = _create_paths(
            build_paths, labeled_paths, self._labeled_edges, self._vertex_to_id
        )

        self._start_vertices = tuple(self._start_vertices)  # copy from iterable
        if build_paths:
            self._predecessors.update(
                (vertex, None)
                for vertex in _iter_start_ids(self._start_vertices, self._vertex_to_id)
            )

        self._calculation_limit = calculation_limit

    def _start(self, *args):
        self._generator = self._traverse(
            self._next_edge_or_vertex,
            self._labeled_edges,
            self._is_tree,
            self._vertex_to_id,
            self._calculation_limit,
            self.paths,
            self._predecessors,
            self._edge_data,
            *args,
        )

    @abstractmethod
    def _traverse(self, *args) -> Iterator[Vertex]:
        """Has to be implemented in sub class"""


class _TraversalWithLabels(Traversal, ABC):
    def __init__(
        self,
        next_edges: NextEdges,
        is_tree: bool,
        vertex_to_id: Optional[VertexToID],
    ) -> None:
        super().__init__(next_edges, True, is_tree, vertex_to_id)


CurrentTraversalClass = TypeVar(
    "CurrentTraversalClass", bound="_TraversalWithOrWithoutLabels"
)


class _TraversalWithOrWithoutLabels(Traversal, ABC):
    def __init__(
        self,
        next_vertices: Optional[NextVertices],
        next_edges: Optional[NextEdges],
        is_tree: bool,
        vertex_to_id: Optional[VertexToID],
    ) -> None:
        if next_vertices is not None:
            if next_edges is not None:
                raise RuntimeError("Both next_vertices and next_edges provided.")
            next_edge_or_vertex = next_vertices
            labeled_edges = False
        else:
            if next_edges is None:
                raise RuntimeError("Neither next_vertices nor next_edges provided.")
            next_edge_or_vertex = next_edges
            labeled_edges = True
        super().__init__(next_edge_or_vertex, labeled_edges, is_tree, vertex_to_id)

        self.visited = None

    def start_from(
        self: CurrentTraversalClass,
        start_vertex: Optional[Vertex] = None,
        *,
        start_vertices: Optional[Iterable[Vertex]] = None,
        build_paths: bool = False,
        labeled_paths: bool = False,
        calculation_limit: Optional[int] = None,
        already_visited: Optional[set] = None,
    ) -> CurrentTraversalClass:
        """
        Starts the traversal at a vertex or a set of vertices and sets parameters.

        :param start_vertex: The vertex the search should start at. If None, provide start_vertices.

        :param start_vertices: The vertices (iterator) the search should start at. Only
            allowed if start_vertex equals None.

        :param build_paths: If true, build paths from some start vertex to each visited
            vertex.

        :param labeled_paths: If true, integrate edge data in generated paths, not only vertices.
            Has to be false in graphs without labeled edges, i.e., if you have given a
            next_vertices function and not a next_edges function.

        :param calculation_limit: If provided, maximal number of vertices to process (read in)
            from your graph. If it is exceeded, a RuntimeError will be raised.

        :param already_visited: If provided, this set is used instead of an internal one to keep
            vertices (resp. their hashable ids from vertex_to_id), that have already been visited.
            This parameter can be used to get online access to the internal bookkeeping of visited
            vertices, or to pre-load vertices that should never be visited, or to provide you own
            way for storing the information that a vertex has already been visited (if you do that
            in your own graph structure instead of a set, visited_vertices needs to provide the
            methods *update* and *__contains__*).

        :return: Traversal, that has been started, e.g., statements like *iter()*, *next()*,
            *for* and the methods "go*" of the Traversal can now be used.
        """

        if not isinstance(type(self), type(Traversal)):
            raise RuntimeError(
                "Method start_from can only be called on a Traversal object."
            )

        self._start_from(
            start_vertex,
            start_vertices,
            build_paths,
            labeled_paths,
            calculation_limit,
        )

        self.visited = _define_visited(
            already_visited,
            _iter_start_ids(self._start_vertices, self._vertex_to_id),
            self._is_tree,
        )
        super()._start()
        return self


class TraversalBreadthFirst(_TraversalWithOrWithoutLabels):
    def __init__(
        self,
        next_vertices: Optional[NextVertices] = None,
        *,
        next_edges: Optional[NextEdges] = None,
        is_tree: bool = False,
        vertex_to_id: Optional[VertexToID] = None,
    ):
        """
        :param next_vertices: See `NextVertices` function. If None, provide next_edges.

        :param next_edges: See `NextEdges` function. Only allowed if next_vertex equals None.

        :param is_tree: bool: If it is sure, that during each traversal run, each vertex can
           be reached only once, is_tree can be set to True. This improves performance,
           but attribute *visited* of the traversal will not be updated during and after
           the traversal.

        :param vertex_to_id: See `VertexToID` function.

        **Algorithm:** Breadth First Search, non-recursive, based on FIFO queue.

        **Properties:** Visits and reports vertices in breadth first order, i.e., with ascending
        depth (edge count of the path with least edges from a start vertex).

        **Input:** Directed graph. Unlabeled or labeled edges. One ore more start vertices.
        Optional calculation limit.

        **Attributes:** When a vertex is *expanded* (traversal calls next_vertices or next_edges)
        or *reported* (an iterator of the traversal returns it), the traversal provides the
        following attributes:

        - **depth:** At this search depth, the reported vertex has been found. It equals the
          length of the created path to the vertex, if path creation is demanded. For the special
          case of TraversalBreadthFirst, it equals the *depth of the vertex* (minimal number
          of edges needed to come to it from a start vertex).

        - **paths:** A container object of class Paths. If path creation has been demanded,
          the container provides the found paths for all vertices visited so far. If labeled edges
          were provided, paths contain them instead of just vertices, if demanded. For the special
          case of TraversalBreadthFirst, all created paths are *shortest paths*, i.e., paths
          with minimal number of edges from a start vertex to their end vertex.

        - **visited**: A collection that contains the vertices (resp. their hashable ids
          from vertex_to_id) that have been visited so far, and the start vertices.
          When a finite graph has been fully traversed, it contains the
          vertices reachable from the start vertices.

        **Inherited methods:** from `Traversal`: `__iter__`, `__next__`, `go_for_vertices_in`,
        `go_to`.
        """

        super().__init__(next_vertices, next_edges, is_tree, vertex_to_id)
        self.depth = None

    def _traverse(self, *args):
        # copy general traversal attributes from parameters into method scope (faster access)
        (
            next_edge_or_vertex,
            labeled_edges,
            is_tree,
            vertex_to_id,
            calculation_limit,
            paths,
            predecessors,
            edge_data,
        ) = args
        # copy traversal specific attributes (set by from_*) into method scope
        visited = self.visited

        # Create booleans (avoid checks with "is")
        edge_data_exists = edge_data is not None

        # two lists used as FIFO queue with just two buckets
        # (using a queue and counting down size of current depth horizon is slower, and
        # creating a new list instead of clear() is also slower)
        to_visit = list(self._start_vertices)
        next_to_visit = list()

        prev_traversal = copy.copy(self)  # copy of self, for keeping previous depth
        self.depth = 1
        prev_traversal.depth = 0

        # Get references of used methods (avoid object resolution)
        visited_add = visited.add
        to_visit_append = to_visit.append
        next_to_visit_append = next_to_visit.append

        if calculation_limit is not None:
            calculation_limit += 1  # allows for limit check by zero check

        while to_visit:
            for vertex in to_visit:
                if calculation_limit and not (
                    calculation_limit := calculation_limit - 1
                ):
                    raise RuntimeError("Number of visited vertices reached limit")

                for edge_or_vertex in next_edge_or_vertex(vertex, prev_traversal):
                    neighbor = edge_or_vertex[0] if labeled_edges else edge_or_vertex

                    if paths or not is_tree:
                        n_id = vertex_to_id(neighbor) if vertex_to_id else neighbor

                        if not is_tree:
                            if n_id in visited:
                                continue
                            visited_add(n_id)

                        if paths:
                            predecessors[n_id] = vertex
                            if edge_data_exists:
                                edge_data[n_id] = edge_or_vertex[1:]

                    yield neighbor
                    next_to_visit_append(neighbor)

            self.depth += 1
            prev_traversal.depth += 1
            to_visit, next_to_visit, to_visit_append, next_to_visit_append = (
                next_to_visit,
                to_visit,
                next_to_visit_append,
                to_visit_append,
            )
            next_to_visit.clear()

    def go_for_depth_range(self, start: int, stop: int) -> VertexIterator:
        """
        For a started traversal, it returns an iterator. During the traversal, the iterator skips
        vertices as long as their depth is lower than *start*. From then on, is reports the found
        vertices. It stops when the reached depth is equal to or higher than *stop*.

        Note: The first vertex with a depth equal or higher than *stop* will be consumed from the
        traversal, but will not be reported, so it is lost (compare *itertools.takewhile*).
        """
        if not isinstance(type(self), type(Traversal)):
            raise RuntimeError(
                "Method go_for_depth_range can only be called "
                + "on a Traversal object."
            )

        # In order to make the above check work, the following generator functionality
        # needs to be encapsulated in a local function
        def my_generator():
            for v in self._generator:
                if self.depth >= start:
                    if self.depth < stop:
                        yield v
                    break
            for v in self._generator:
                if self.depth >= stop:
                    break
                yield v

        return my_generator()


class TraversalDepthFirst(_TraversalWithOrWithoutLabels):
    """
    :param next_vertices: See `NextVertices` function. If None, provide next_edges.

    :param next_edges: See `NextEdges` function. Only allowed if next_vertex equals None.

    :param is_tree: bool: If it is sure, that during each traversal run, each vertex can
       be reached only once, is_tree can be set to True. This improves performance,
       but attribute *visited* of the traversal will not be updated during and after
       the traversal.

    :param vertex_to_id: See `VertexToID` function.

    **Algorithm:** Depth First Search ("BFS"), non-recursive, based on stack.

    **Properties:** Follows edges to new vertices as long as possible, and goes back a step and
    follows further edges that start at some visited vertex only if necessary to come to new
    vertices.

    **Input:** Directed graph. One ore more start vertices. Vertices must be hashable, or hashable
    id can be provided. Unlabeled or labeled edges. Optional calculation limit.

    **Attributes:** When a vertex is *expanded* (traversal calls next_vertices or next_edges)
    or *reported* (an iterator of the traversal returns it), the traversal provides the
    following attributes:

    - **depth:** At this search depth, the reported vertex has been found. It equals the length of
      the created path to the vertex, if path creation is demanded. Note: The search depth does not
      need to be the depth of the vertex.

    - **paths:** A container object of class Paths. If path creation has been demanded,
      the container provides the found paths for all vertices visited so far. If labeled edges
      were provided, paths contain them instead of just vertices, if demanded.

    - **visited:** A collection that contains the vertices that have been visited so far, and the
      start vertices. When a finite graph has been fully traversed, it contains the vertices
      reachable from the start vertices.

    **Inherited methods:** from `Traversal`: `__iter__`, `__next__`, `go_for_vertices_in`,
    `go_to`.
    """

    def __init__(
        self,
        next_vertices: Optional[NextVertices] = None,
        *,
        next_edges: Optional[NextEdges] = None,
        is_tree: bool = False,
        vertex_to_id: Optional[VertexToID] = None,
    ) -> None:
        super().__init__(next_vertices, next_edges, is_tree, vertex_to_id)
        self.depth = None

    def _traverse(self, *args):
        # copy general traversal attributes from parameters into method scope (faster access)
        (
            next_edge_or_vertex,
            labeled_edges,
            is_tree,
            vertex_to_id,
            calculation_limit,
            paths,
            predecessors,
            edge_data,
        ) = args

        # copy traversal specific attributes (set by from_*) into method scope
        visited = self.visited

        # Create booleans (avoid checks with "is")
        edge_data_exists = edge_data is not None

        depth = 0
        to_visit = list(self._start_vertices)  # list used as stack

        # Get references of methods (avoid object resolution)
        to_visit_append = to_visit.append
        visited_add = visited.add

        if calculation_limit is not None:
            calculation_limit += 1  # allows for limit check by zero check

        while to_visit:
            vertex = to_visit.pop()  # visit first added vertex first

            if (
                not is_tree and vertex is None
            ):  # Reached marker: end of vertices in same depth
                depth -= 1
                continue

            if depth:
                if is_tree:
                    self.depth = depth
                    yield vertex
                else:
                    n_id = vertex_to_id(vertex) if vertex_to_id else vertex
                    if n_id in visited:
                        continue
                    visited_add(n_id)
                    self.depth = depth
                    yield vertex
                    to_visit_append(
                        None
                    )  # Marker: reached by pop means leaving the vertex
            else:  # start vertices can be expanded (but not yielded) although being visited
                self.depth = depth
                if not is_tree:
                    to_visit_append(
                        None
                    )  # Marker: reached by pop means leaving the vertex

            if calculation_limit and not (calculation_limit := calculation_limit - 1):
                raise RuntimeError("Number of visited vertices reached limit")

            for edge_or_vertex in next_edge_or_vertex(vertex, self):
                neighbor = edge_or_vertex[0] if labeled_edges else edge_or_vertex
                if paths:
                    n_id = vertex_to_id(neighbor) if vertex_to_id else neighbor

                    if not is_tree and n_id in visited:
                        continue

                    # We have to store the predecessor here, because at time of visit, it is
                    # already lost. And we cannot yield here, because only the first of the
                    # neighbors will indeed be visited next.
                    # But since the visiting order is defined by a stack we know that for each
                    # vertex, the predecessor stored last is the edge visited first, and after
                    # that no other predecessor can be stored for that vertex.
                    predecessors[n_id] = vertex
                    if edge_data_exists:
                        edge_data[n_id] = edge_or_vertex[1:]

                to_visit_append(neighbor)
            depth += 1


class TraversalTopologicalSort(_TraversalWithOrWithoutLabels):
    """
    :param next_vertices: See `NextVertices` function. If None, provide next_edges.

    :param next_edges: See `NextEdges` function. Only allowed if next_vertex equals None.

    :param is_tree: bool: If it is sure, that during each traversal run, each vertex can
       be reached only once, is_tree can be set to True. This improves performance,
       but attribute *visited* of the traversal will not be updated during and after
       the traversal.

    :param vertex_to_id: See `VertexToID` function.

    **Algorithm:** Topological Search, non-recursive, based on stack.

    **Properties:** Vertices are reported in topological ordering, i.e. a linear ordering of
    the vertices such that for every directed edge uv from vertex u to vertex v ("u depends on
    v"), v comes before u in the ordering. If the graph contains a cycle that can be reached
    within the sorting process, a RuntimeError exception is raised and a cyclic path from a
    starting vertex is provided.

    **Input:** Directed graph. One ore more start vertices. Vertices must be hashable, or hashable
    id can be provided. Unlabeled or labeled edges. Optional calculation limit. Edge from vertex
    u to vertex v means, u "depends" from v.

    **Attributes:** When a vertex is *expanded* (traversal calls next_vertices or next_edges)
    or *reported* (an iterator of the traversal returns it), the traversal provides the
    following attributes:

    - **depth:** At this search depth, the reported vertex has been found. It equals the length of
      the created path to the vertex, if path creation is demanded. Note: The search depth does not
      need to be the depth of the vertex.

    - **paths:** A container object of class Paths. If path creation has been demanded,
      the container provides the found paths for all vertices visited so far. If labeled edges
      were provided, paths contain them instead of just vertices, if demanded.

    - **visited:** A collection that contains the vertices that have been visited so far, and the
      start vertices. When a finite graph has been fully traversed, it contains the vertices
      reachable from the start vertices.

      Note: TraversalTopologicalSort often visits vertices long before it reports them.
      Before it can report a vertex, it needs to report all the vertices, that the vertex depends
      on (directly or indirectly).

    If the graph contains a cycle that can be reached within the sorting process, a RuntimeError
    exception is raised, and the traversal provides the following additional attribute:

    - **cycle_from_start:** A cyclic path from a starting vertex.

    **Inherited methods:** from `Traversal`: `__iter__`, `__next__`, `go_for_vertices_in`,
    `go_to`.
    """

    def __init__(
        self,
        next_vertices: Optional[NextVertices] = None,
        *,
        next_edges: Optional[NextEdges] = None,
        is_tree: bool = False,
        vertex_to_id: Optional[VertexToID] = None,
    ) -> None:
        super().__init__(next_vertices, next_edges, is_tree, vertex_to_id)
        self.depth = None
        self.cycle_from_start = None

    def _traverse(self, *args):
        # copy general traversal attributes from parameters into method scope (faster access)
        (
            next_edge_or_vertex,
            labeled_edges,
            is_tree,
            vertex_to_id,
            calculation_limit,
            paths,
            predecessors,
            edge_data,
        ) = args

        # Copy traversal specific attributes (set by from_*) into method scope
        visited = self.visited

        # Create booleans (avoid checks with "is")
        edge_data_exists = edge_data is not None

        # Initialization
        self.depth = 0

        if calculation_limit is not None:
            calculation_limit += 1  # allows for limit check by zero check

        # Get references of methods (avoid object resolution)
        visited_add = visited.add

        # Two separate implementations for the cases is_tree and not is_tree that follow
        # different concepts, because a combined approach makes both cases significantly
        # slower
        if is_tree:
            # Initialization
            to_visit = list(
                itertools.chain.from_iterable(
                    itertools.zip_longest(self._start_vertices, [])
                )
            )

            # Get references of methods (avoid object resolution)
            to_visit_extend = to_visit.extend
            to_visit_pop = to_visit.pop

            while to_visit:
                vertex = to_visit_pop()  # visit last added vertex first

                if vertex is not None:  # We "leave" and report the vertex
                    self.depth -= 1
                    yield vertex
                    continue

                # We "expand" the vertex
                vertex = to_visit[-1]

                if calculation_limit and not (
                    calculation_limit := calculation_limit - 1
                ):
                    raise RuntimeError("Number of visited vertices reached limit")

                for edge_or_vertex in next_edge_or_vertex(vertex, self):
                    neighbor = edge_or_vertex[0] if labeled_edges else edge_or_vertex
                    n_id = vertex_to_id(neighbor) if vertex_to_id else neighbor

                    if paths:
                        # We have to store the predecessor here, because at time of visit, it is
                        # already lost. And we cannot yield here, because only the first of the
                        # neighbors will indeed be visited next.
                        # But since the visiting order is defined by a stack we know that for each
                        # vertex, the predecessor stored last is the edge visited first, and after
                        # that no other predecessor can be stored for that vertex.
                        predecessors[n_id] = vertex
                        if edge_data_exists:
                            edge_data[n_id] = edge_or_vertex[1:]

                    to_visit_extend(
                        (neighbor, None)
                    )  # Vertex, and marker "not expanded so far"

                self.depth += 1
        else:
            # Initialization
            to_visit = list(self._start_vertices)  # list used as stack of todos
            trace_set = set()  # set of vertices along the current path

            # Get references of methods (avoid object resolution)
            to_visit_append = to_visit.append
            to_visit_pop = to_visit.pop
            trace_set_add = trace_set.add
            trace_set_discard = trace_set.discard

            while to_visit:
                vertex = to_visit[-1]  # visit or report last added vertex first
                v_id = vertex_to_id(vertex) if vertex_to_id else vertex

                if v_id in trace_set:
                    # Back to trace, from visits/reports of further vertices, that trace vertices
                    # depend on: We "leave" and report the head vertex of the trace.
                    self.depth -= 1
                    to_visit_pop()
                    trace_set_discard(v_id)
                    yield vertex
                    continue

                # Visit vertex, but not when marked as already visited.
                # Ignore this precondition for trees and start vertices.
                if self.depth > 0:
                    if v_id in visited:
                        to_visit_pop()
                        continue
                    visited_add(v_id)

                # Now, vertex belongs to trace from start. As long as this is so, seeing it as
                # neighbor would be a cycle.
                trace_set_add(v_id)

                # We "expand" the vertex
                if calculation_limit and not (
                    calculation_limit := calculation_limit - 1
                ):
                    raise RuntimeError("Number of visited vertices reached limit")

                for edge_or_vertex in next_edge_or_vertex(vertex, self):
                    neighbor = edge_or_vertex[0] if labeled_edges else edge_or_vertex
                    n_id = vertex_to_id(neighbor) if vertex_to_id else neighbor

                    if n_id in trace_set:
                        # We found a dependency (edge) back to a vertex, whose dependencies
                        # we are currently following (trace). We build and report this trace.
                        trace = list()
                        for vertex in to_visit:
                            v_id = vertex_to_id(vertex) if vertex_to_id else vertex
                            if v_id in trace_set:
                                trace.append(vertex)
                        trace.append(neighbor)
                        self.cycle_from_start = trace
                        raise RuntimeError("Graph contains cycle")

                    if n_id in visited:
                        continue

                    if paths:
                        # We have to store the predecessor here, because at time of visit, it is
                        # already lost. And we cannot yield here, because only the first of the
                        # neighbors will indeed be visited next.
                        # But since the visiting order is defined by a stack we know that for each
                        # vertex, the predecessor stored last is the edge visited first, and after
                        # that no other predecessor can be stored for that vertex.
                        predecessors[n_id] = vertex
                        if edge_data_exists:
                            edge_data[n_id] = edge_or_vertex[1:]

                    to_visit_append(neighbor)

                self.depth += 1


# --------------- traversal strategies for weighted graphs -------------


class TraversalShortestPaths(_TraversalWithLabels):
    """
    :param next_edges: See `NextEdges` function.

    :param is_tree: bool: If it is sure, that during each traversal run, each vertex can
       be reached only once, is_tree can be set to True. This improves performance,
       but attribute *distances* of the traversal will not be updated during and after
       the traversal.

    :param vertex_to_id: See `VertexToID` function.

    **Algorithm:** Shortest paths algorithm of Dijkstra, non-recursive, based on heap.

    **Properties:** Vertices are visited and reported ordered by increasing distance (minimally
    necessary sum of edge weights) from a start vertex. Each vertex is visited only once.

    **Input:** Directed graph. One ore more start vertices. Vertices must be hashable, or hashable
    id can be provided. Labeled edges, first field (after to_vertex) is weight, weight needs to be
    non-negative. Optional calculation limit.

    **Attributes:** When a vertex is *expanded* (traversal calls next_vertices or next_edges)
    or *reported* (an iterator of the traversal returns it), the traversal provides the
    following attributes:

    - **distance:** The length of the shortest path (sum of edge weights) from a start vertex
      to the visited vertex

    - **depth:** At this search depth, the reported vertex has been found. It equals the edge count
      of the created path to the vertex, if path creation is demanded. Note: The search depth does
      not need to be the depth of the vertex.

    - **paths:** A container object of class Paths. If path creation has been demanded, the
      container provides a shortest path for all vertices visited so far. If labeled edges were
      provided, paths contain them instead of just vertices, if demanded.

    - **distances:** A dictionary. For the vertices that have already been visited, it contains
      their distance.

      Note: Typically, it contains values for some other vertices, too. These
      might not be final and could change until the respective vertex is visited.

      When a finite graph has been fully traversed, it contains the distances of all vertices
      that are reachable from the start vertices.

    **Inherited methods:** from `Traversal`: `__iter__`, `__next__`, `go_for_vertices_in`,
    `go_to`.
    """

    def __init__(
        self,
        next_edges: NextEdges,
        *,
        is_tree: bool = False,
        vertex_to_id: Optional[VertexToID] = None,
    ) -> None:
        super().__init__(next_edges, is_tree, vertex_to_id)
        self.distance = None
        self.depth = None
        self.distances = None

    def start_from(
        self,
        start_vertex: Optional[Vertex] = None,
        *,
        start_vertices: Optional[Iterable[Vertex]] = None,
        build_paths: bool = False,
        labeled_paths: bool = False,
        calculation_limit: Optional[int] = None,
        known_distances: Optional[dict[Hashable, Real]] = None,
    ) -> TraversalShortestPaths:
        """
        :param start_vertex: The vertex the search should start at. If None, provide start_vertices.

        :param start_vertices: The set of vertices (iterator) the search should start at. Only
            allowed if start_vertex equals None.

        :param build_paths: If true, build paths from start vertices for each visited vertex.

        :param labeled_paths: If true, integrate edge data in generated paths, not only vertices.

        :param calculation_limit: If provided, maximal number of vertices to process (read in)
            from your graph. If it is exceeded, a RuntimeError will be raised.

        :param known_distances: If provided, this dict is used instead of an internal one to keep
            the distances of vertices that have already been visited (resp. their hashable ids
            from vertex_to_id is used as key) from some start vertex. For vertices without known
            distance, it must yield float('infinity'). The internal default implementation uses
            a collections.defaultdict. Typical use cases are:
            1) pre-loading known distances of vertices, and the vertices should not be visited if
            no smaller distance is found during the traversal, or 2) getting online access to the
            internal bookkeeping of visited vertices and their distances, or 3) for providing your
            own way for storing the distance of a vertex that has already been visited (if you do
            that in your own graph structure instead of a dict, visited_vertices needs to provide
            the methods __contains__, __getitem__ and __setitem__).

        :return: Traversal, that has been started, e.g., the methods go* can now be used.
        """

        if not isinstance(type(self), type(Traversal)):
            raise RuntimeError(
                "Method start_from can only be called on a Traversal object."
            )

        self._start_from(
            start_vertex,
            start_vertices,
            build_paths,
            labeled_paths,
            calculation_limit,
        )
        self.distance = None
        self.depth = None
        self.distances = known_distances
        super()._start()
        return self

    def _traverse(self, *args):
        # copy general traversal attributes from parameters into method scope (faster access)
        (
            next_edges,
            _labeled_edges,
            is_tree,
            vertex_to_id,
            calculation_limit,
            paths,
            predecessors,
            edge_data,
        ) = args

        # Create booleans (avoid checks with "is")
        edge_data_exists = edge_data is not None

        # At start, most of the distances from a vertex to a start vertex are not known. If
        # accessed for comparison for possibly better distances, infinity is used, if no other
        # value is given.
        if self.distances is None:
            infinity = float("infinity")
            self.distances = collections.defaultdict(lambda: infinity)
        distances = self.distances

        # So far, the start vertices are to be visited. Each has distance 0 from a start vertex
        # (itself), if not defined otherwise, an edge count of 0, and a path, if one is required,
        # consisting of only the vertex itself.
        start_vertices_and_ids = tuple(
            (vertex, vertex_to_id(vertex) if vertex_to_id else vertex)
            for vertex in self._start_vertices
        )

        for vertex, vertex_id in start_vertices_and_ids:
            distances.setdefault(vertex_id, 0)  # set to 0, if not already set

        # Unique number, that prevents heapq from sorting by vertices in case of a tie in the sort
        # field, because vertices do not need to be pairwise comparable. The integers from -5 to 256
        # are used first, because they are internalized (pre-calculated, and thus fastest). We count
        # downwards like we do in A* search. There, it is preferable, because a LIFO behavior makes
        # A* often faster. Here, we do it simply to do it the same way.
        unique_no = itertools.count(256, -1)
        to_visit = [  # used as collection.heapq of tuples, lowest distance first
            (distances[vertex_id], next(unique_no), vertex, 0)
            for vertex, vertex_id in start_vertices_and_ids
        ]
        heapify(to_visit)

        if calculation_limit is not None:
            calculation_limit += 1  # allows for limit check by zero check

        while to_visit:
            # Visit path with lowest distance first
            path_weight, _, vertex, path_edge_count = heappop(to_visit)

            # A vertex can get added to the heap multiple times. We want to process it only once,
            # the first time it is removed from the heap, because this is the case with the least
            # distance from start.
            if not is_tree:
                v_id = vertex_to_id(vertex) if vertex_to_id else vertex
                if path_weight > distances[v_id]:
                    continue
                # (Allow garbage collector to free distance value if nowhere else needed any more)
                distances[v_id] = 0

            # Export traversal data to traversal attributes
            self.distance = path_weight
            self.depth = path_edge_count

            # We now know the distance of the vertex, so we report it.
            if path_edge_count > 0:  # do not yield start vertex
                yield vertex

            if calculation_limit and not (calculation_limit := calculation_limit - 1):
                raise RuntimeError("Number of visited vertices reached limit")

            n_path_edge_count = path_edge_count + 1
            for edge in next_edges(vertex, self):
                neighbor, weight = edge[0], edge[1]
                n_path_weight = path_weight + weight

                # If, so far, we have not found a shorter path to the neighbor than the new one
                # that ends with the edge, this path is a candidate for a shortest path to the
                # neighbor. We push it to the heap.
                if paths or not is_tree:
                    n_id = vertex_to_id(neighbor) if vertex_to_id else neighbor

                    if not is_tree:
                        if distances[n_id] <= n_path_weight:
                            continue
                        distances[n_id] = n_path_weight

                    # If we are to generate a path, we have to do it here, since the edge we have
                    # to add to the path prefix is not stored on the heap
                    if paths:
                        predecessors[n_id] = vertex
                        if edge_data_exists:
                            edge_data[n_id] = edge[1:]

                heappush(
                    to_visit,
                    (
                        n_path_weight,
                        next(unique_no),
                        neighbor,
                        n_path_edge_count,
                    ),
                )

    def go_for_distance_range(self, start: int, stop: int) -> VertexIterator:
        """
        For a started traversal, it returns an iterator.  During the traversal, the iterator skips
        vertices as long as their distance is lower than *start*. From then on, is reports the found
        vertices. It stops when the reached distance is equal to or higher than *stop*.

        Note: The first vertex with a distance equal or higher than stop will be consumed from the
        traversal, but will not be reported, so it is lost (compare itertools.takewhile).
        """
        if not isinstance(type(self), type(Traversal)):
            raise RuntimeError(
                "Method go_for_distance_range can only be called "
                + "on a Traversal object."
            )

        # In order to make the above check work, the following generator functionality
        # needs to be encapsulated in a local function
        def my_generator():
            for v in self._generator:
                if self.distance >= start:
                    if self.distance < stop:
                        yield v
                    break
            for v in self._generator:
                if self.distance >= stop:
                    break
                yield v

        return my_generator()


class TraversalAStar(_TraversalWithLabels):
    """
    :param next_edges: See `NextEdges` function.

    :param is_tree: bool: If it is sure, that during each traversal run, each vertex can
       be reached only once, is_tree can be set to True. This improves performance,
       but attribute *path_length_guesses* of the traversal will not be updated during the
       traversal.

    :param vertex_to_id: See `VertexToID` function.

    **Algorithm:** The search algorithm A*, non-recursive, based on heap.

    **Input:** Weighted directed graph. Only non-negative edge weights. One or more start
    vertices. Vertices must be hashable, but need not to be comparable. A heuristic function that
    estimates the cost of the cheapest path from a given vertex to the goal (resp. to any of your
    goal vertices, if you have more than one), and never overestimates the actual needed costs
    ("admissible heuristic function"). Optionally, a calculation limit.

    **Properties:** Vertices are visited and reported ordered by increasing path length (sum of
    edge weights) of the shortest path from a start vertex to the visited vertex that have been
    found so far (!).

    When the goal is visited, the reported path is a shortest path from start to goal and
    the reported length is the distance of the goal from start.

    In case the used heuristic function is consistent, i.e. following an edge from one vertex to
    another never reduces the estimated costs to get to the goal by more than the weight of the
    edge, further guarantees hold: Each vertex is only visited once. And for each visited vertex,
    the yielded path length and edge count (and optionally, the path) are the data of the
    shortest existing path from start (not only from the shortest path found so far).

    **Attributes:** When a vertex is *expanded* (traversal calls next_vertices or next_edges)
    or *reported* (an iterator of the traversal returns it), the traversal provides the
    following attributes:

    - **path_length:** Length of the found path to the vertex (for the goal vertex: a shortest path)

    - **depth:** At this search depth, the reported vertex has been found. It equals the edge count
      of the created path to the vertex, if path creation is demanded. Note: The search depth does
      not need to be the depth of the vertex.

    - **paths:** A container object of class Paths.  If path creation has been demanded,
      the container provides the generated paths for all vertices visited so far. If labeled edges
      were provided, paths contain them instead of just vertices, if demanded.

    **Inherited methods:** from `Traversal`: `__iter__`, `__next__`, `go_for_vertices_in`,
    `go_to`.
    """

    def __init__(
        self,
        next_edges: NextEdges,
        *,
        is_tree: bool = False,
        vertex_to_id: Optional[VertexToID] = None,
    ) -> None:
        super().__init__(next_edges, is_tree, vertex_to_id)
        self.path_length = None
        self.depth = None

        self._heuristic = None
        self._known_distances = None
        self._known_path_length_guesses = None

    def start_from(
        self,
        heuristic: Callable[[Vertex], Real],
        start_vertex: Optional[Vertex] = None,
        *,
        start_vertices: Optional[Iterable[Vertex]] = None,
        build_paths: bool = False,
        labeled_paths: bool = False,
        calculation_limit: Optional[int] = None,
        known_distances: Optional[dict[Hashable, Real]] = None,
        known_path_length_guesses: Optional[dict[Hashable, Real]] = None,
    ) -> TraversalAStar:
        """
        :param heuristic: The admissible and consistent heuristic function that estimates the cost
            of the cheapest path from a given vertex to the goal (resp. one of the goals).

        :param start_vertex: The vertex the search should start at. Provide either start_vertex
            or start_vertices, but not both.

        :param start_vertices: The set of vertices (iterator) the search should start at. Provide
            either start_vertex or start_vertices, but not both.

        :param build_paths: If true, build paths from start vertices for each visited vertex.

        :param labeled_paths: If true, integrate edge data in generated paths, not only vertices.

        :param calculation_limit: If provided, maximal number of vertices to process (read in)
            from your graph. If it is exceeded, a RuntimeError will be raised.

        :param known_distances: If provided, this dict is used instead of an internal one to keep
            the distances of vertices that have already been visited (resp. their hashable ids
            from vertex_to_id is used as key) from some start vertex. For vertices without known
            distance, it must yield float('infinity'). The internal default implementation uses
            a collections.defaultdict. Typical use cases are:
            1) pre-loading known distances of vertices, and the vertices should not be visited if
            no smaller distance is found during the traversal, or 2) getting online access to the
            internal bookkeeping of visited vertices and their distances, or 3) for providing your
            own way for storing the distance of a vertex that has already been visited (if you do
            that in your own graph structure instead of a dict, visited_vertices needs to provide
            the methods __contains__, __getitem__ and __setitem__).

        :param known_path_length_guesses: Like known_distances, but for keeping the sum
            distance+heuristic for vertices.

        :return: Traversal, that has been started, e.g., the methods go* can now be used.
        """

        if not isinstance(type(self), type(Traversal)):
            raise RuntimeError(
                "Method start_from can only be called on a Traversal object."
            )

        self._start_from(
            start_vertex,
            start_vertices,
            build_paths,
            labeled_paths,
            calculation_limit,
        )
        self.path_length = None
        self.depth = None
        self._heuristic = heuristic
        self._known_distances = known_distances
        self._known_path_length_guesses = known_path_length_guesses
        super()._start()
        return self

    def _traverse(self, *args):
        # copy general traversal attributes from parameters into method scope (faster access)
        (
            next_edges,
            _labeled_edges,
            is_tree,
            vertex_to_id,
            calculation_limit,
            paths,
            predecessors,
            edge_data,
        ) = args

        # copy traversal specific attributes from parameters into method scope (faster access)
        heuristic = self._heuristic

        # Create booleans (avoid checks with "is")
        edge_data_exists = edge_data is not None

        # At start, most of the distances from a vertex to a start vertex are not known. If
        # accessed for comparison for possibly better distances, infinity is used, if no other
        # value is given.
        distances = self._known_distances
        if distances is None:
            infinity = float("infinity")
            distances = collections.defaultdict(lambda: infinity)

        path_length_guesses = self._known_path_length_guesses
        if path_length_guesses is None:
            infinity = float("infinity")
            path_length_guesses = collections.defaultdict(lambda: infinity)

        # So far, the start vertices are to be visited. Each has distance 0 from a start vertex
        # (itself), if not defined otherwise, an edge count of 0, and a path, if one is required,
        # consisting of only the vertex itself.
        start_vertices_and_ids = tuple(
            (vertex, vertex_to_id(vertex) if vertex_to_id else vertex)
            for vertex in self._start_vertices
        )

        for vertex, vertex_id in start_vertices_and_ids:
            distances.setdefault(vertex_id, 0)  # set to 0, if not already set
            path_length_guesses.setdefault(
                vertex_id, distances[vertex_id] + heuristic(vertex)
            )

        # Unique number, that prevents heapq from sorting by vertices in case of a tie in the sort
        # field, because vertices do not need to be pairwise comparable. The numbers are generated
        # in decreasing order to make the min heap behave like a LIFO queue in case of ties. The
        # integers from -5 to 256 are used first, because they are internalized (pre-calculated,
        # and thus fastest).
        unique_no = itertools.count(256, -1)
        to_visit = [  # used as collection.heapq of tuples, lowest distance first
            (path_length_guesses[vertex_id], next(unique_no), vertex, 0)
            for vertex, vertex_id in start_vertices_and_ids
        ]
        heapify(to_visit)

        if calculation_limit is not None:
            calculation_limit += 1  # allows for limit check by zero check

        while to_visit:
            # Visit path with lowest distance first
            path_length_guess, _, vertex, path_edge_count = heappop(to_visit)

            # A vertex can get added to the heap multiple times.
            # For consistent heuristics: We want to process the vertex only once, the first time it
            # is removed from the heap, because this is the case with the least distance estimation.
            # If the heuristic is not consistent: Only when the new distance estimation is better
            # than the best found so far, we want to process the vertex again.
            v_id = vertex_to_id(vertex) if vertex_to_id else vertex
            if not is_tree and path_length_guess > path_length_guesses[v_id]:
                continue

            path_weight = distances[v_id]

            # Export traversal data to traversal attributes
            self.path_length = path_weight
            self.depth = path_edge_count

            # We now know the distance of the vertex, so we report it.
            if path_edge_count > 0:  # do not yield start vertex
                yield vertex

            if calculation_limit and not (calculation_limit := calculation_limit - 1):
                raise RuntimeError("Number of visited vertices reached limit")

            n_path_edge_count = path_edge_count + 1
            for edge in next_edges(vertex, self):
                neighbor, weight = edge[0:2]
                n_path_weight = path_weight + weight

                # If the new found path to the neighbor is longer than the shortest found so far,
                # we can safely ignore the new path. Otherwise, it is a new candidate for a
                # shortest path to the neighbor, and we push it to the heap.
                n_id = vertex_to_id(neighbor) if vertex_to_id else neighbor
                if not is_tree and distances[n_id] <= n_path_weight:
                    continue

                distances[n_id] = n_path_weight

                # If we are to generate a path, we have to do it here, since the edge we have
                # to add to the path prefix is not stored on the heap
                if paths:
                    predecessors[n_id] = vertex
                    if edge_data_exists:
                        edge_data[n_id] = edge[1:]

                n_guess = n_path_weight + heuristic(neighbor)
                if not is_tree:
                    path_length_guesses[n_id] = n_guess
                heappush(
                    to_visit,
                    (n_guess, next(unique_no), neighbor, n_path_edge_count),
                )


class TraversalMinimumSpanningTree(_TraversalWithLabels):
    """
    :param next_edges: See `NextEdges` function.

    :param vertex_to_id: See `VertexToID` function.

    **Algorithm:**  Minimum spanning tree ("MST") algorithm of Jarnik, Prim, Dijkstra.
    Non-recursive, based on heap. A so-called *tie breaker* is implemented, that
    prioritizes edges that have been found more recently about edges that have been
    found earlier. This is a typical choice that often improves search performance.

    **Properties:** Only edges of the MST from start vertices are reported. Each vertex is visited
    only once.

    **Input:** Weighted undirected graph, given as directed edges with the same weight in both
    directions. One or more start vertices (e.g. for components in unconnected graphs).
    Labeled edges, first field (after to_vertex) is weight. Optional calculation limit.

    **Attributes:** When a vertex is *expanded* (traversal calls next_vertices or next_edges)
    or an edge is *reported* (an iterator of the traversal returns the vertex it leads to),
    the traversal provides the following attributes:

    - **edge:** Tuple of from_vertex, to_vertex, the weight of the edge, and additional data
      you have provided with the edge

    - **paths:** A container object of class Paths. If path creation has been demanded,
      the container provides a path within the MST from a start vertex to each of the vertices
      visited so far. If labeled edges were provided, paths contain them instead of just
      vertices, if demanded.

    **Inherited methods:** from `Traversal`: `__iter__`, `__next__`, `go_for_vertices_in`,
    `go_to`.
    """

    def __init__(
        self,
        next_edges: NextEdges,
        *,
        vertex_to_id: Optional[VertexToID] = None,
    ) -> None:
        super().__init__(next_edges, False, vertex_to_id)
        self.edge = None

    def start_from(
        self,
        start_vertex: Optional[Vertex] = None,
        *,
        start_vertices: Optional[Iterable[Vertex]] = None,
        build_paths: bool = False,
        labeled_paths: bool = False,
        calculation_limit: Optional[int] = None,
    ) -> TraversalMinimumSpanningTree:
        """
        :param start_vertex: The vertex the search should start at. If None, provide start_vertices.

        :param start_vertices: The set of vertices (iterator) the search should start at. Only
            allowed if start_vertex equals None. Leads to a result consisting of several trees
            that are only connected if the start vertices are connected.

        :param build_paths: If true, build paths from start vertices for each visited vertex.

        :param labeled_paths: If true, integrate edge data in generated paths, not only vertices.

        :param calculation_limit: If provided, maximal number of vertices to process (read in)
            from your graph. If it is exceeded, a RuntimeError will be raised.

        :return: Traversal, that has been started, e.g., the methods go* can now be used.
        """

        if not isinstance(type(self), type(Traversal)):
            raise RuntimeError(
                "Method start_from can only be called on a Traversal object."
            )

        self._start_from(
            start_vertex,
            start_vertices,
            build_paths,
            labeled_paths,
            calculation_limit,
        )
        self.edge = None
        super()._start()
        return self

    def _traverse(self, *args):
        # copy general traversal attributes from parameters into method scope (faster access)
        (
            next_edges,
            _labeled_edges,
            is_tree,
            vertex_to_id,
            calculation_limit,
            paths,
            predecessors,
            edge_data,
        ) = args

        # Create booleans (avoid checks with "is")
        edge_data_exists = edge_data is not None

        # At start, only the start vertices are regarded as visited
        visited = set(_iter_start_ids(self._start_vertices, vertex_to_id))

        # Check if we we already go over the calculation limit when we evaluate the edges from
        # start vertices ("expanding the start vertices"). This avoids a step by step check that
        # slows down the to_visit loop for large sets of start vertices.
        # Note: A calculation limit below 0 leads nowhere ever to an exception. Also here.
        if calculation_limit is not None and calculation_limit >= 0:
            if (calculation_limit := calculation_limit - len(self._start_vertices)) < 0:
                raise RuntimeError("Number of visited vertices reached limit")

        # So far, the edges from the start vertices are to be visited as candidates for edges of a
        # MST. (Unique number prevents heapq from sorting by (possibly not comparable) fields)
        unique_no = itertools.count()
        to_visit = [  # used as collection.heapq, lowest edge weight first
            (edge[1], next(unique_no), vertex, edge)
            for vertex in self._start_vertices
            for edge in next_edges(vertex, self)
        ]
        heapify(to_visit)

        # Get references of methods (avoid object resolution)
        visited_add = visited.add

        if calculation_limit is not None:
            calculation_limit += 1  # allows for limit check by zero check

        while to_visit:
            # Visit edge with lowest weight first
            _weight, _, vertex, to_edge = heappop(to_visit)
            to_vertex = to_edge[0]

            # A vertex can get added to the heap multiple times, as end vertex of several edges. We
            # want to process it only once, as end vertex of a MST edge.
            to_id = vertex_to_id(to_vertex) if vertex_to_id else to_vertex
            if to_id in visited:
                continue

            # The shortest edge from a visited vertex that leads to a vertex not visited so far,
            # must be an edge of the MST.
            visited_add(to_id)

            if paths:
                predecessors[to_id] = vertex
                if edge_data_exists:
                    _, *l_edge_data = to_edge
                    edge_data[to_id] = l_edge_data

            # Export traversal data to traversal attribute and report vertex
            self.edge = (vertex,) + to_edge
            yield to_vertex

            if calculation_limit and not (calculation_limit := calculation_limit - 1):
                raise RuntimeError("Number of visited vertices reached limit")

            for n_to_edge in next_edges(to_vertex, self):
                n_to_vertex, n_weight = n_to_edge[0], n_to_edge[1]
                # If the edge leads to a vertex that is, so far, not reached by edges of the MST,
                # it is a candidate for a MST edge. We push it to the heap.
                n_to_id = vertex_to_id(n_to_vertex) if vertex_to_id else n_to_vertex
                if n_to_id not in visited:
                    heappush(
                        to_visit,
                        (n_weight, next(unique_no), to_vertex, n_to_edge),
                    )



import collections
import nographs as nog


# --- Utility functions ---
def edges(next_edges_or_vertices, vertices):
    # Test functionality of a next_edges or next_vertices function by calling
    # it for all vertices and return the relationship between inputs and
    # outputs in a dict.
    return dict((vertex, next_edges_or_vertices(vertex, None))
                for vertex in vertices)


# --- Tests ---


class ExamplesFromDocs:
    """
    >>> def spiral_graph(i, _):
    ...     j = (i + i // 6) % 6
    ...     yield i + 1, j * 2 + 1
    ...
    ...     if i % 2 == 0:
    ...         yield i + 6, 7 - j
    ...
    ...     else:
    ...         if i > 5:
    ...             yield i - 6, 1
    >>> traversal = nog.TraversalShortestPaths(spiral_graph).start_from(0, build_paths=True)
    >>> traversal.go_to(5)
    5
    >>> traversal.distance
    24
    >>> traversal.paths[5]
    (0, 1, 2, 3, 4, 10, 16, 17, 11, 5)

    -- next_from_edge_dict for weighted edges given as dict --
    >>> edges_weighted_dict = {
    ...    'A': (('B', 1), ('C', 2)),
    ...    'B': (('C', 5), ('D', 5)),
    ...    'C': (('D', 2), ('E', 1), ('F', 5)),
    ...    'D': (('F', 2), ('G', 5)),
    ...    'E': (('F', 5), ('B', 1)),
    ...    'F': (('G', 2),)
    ... }

    >>> next_vertices = nog.adapt_edge_index(edges_weighted_dict,add_inverted=False,labeled=True)
    >>> edges(next_vertices, edges_weighted_dict.keys()) == edges_weighted_dict
    True

    >>> next_vertices = nog.adapt_edge_index(edges_weighted_dict,add_inverted=True,labeled=True)
    >>> edges(next_vertices, edges_weighted_dict.keys())  # doctest: +NORMALIZE_WHITESPACE
    {'A': [('B', 1), ('C', 2)], 'B': [('A', 1), ('C', 5), ('D', 5), ('E', 1)],
     'C': [('A', 2), ('B', 5), ('D', 2), ('E', 1), ('F', 5)],
     'D': [('B', 5), ('C', 2), ('F', 2), ('G', 5)],
     'E': [('C', 1), ('F', 5), ('B', 1)],
     'F': [('C', 5), ('D', 2), ('E', 5), ('G', 2)]}


    -- next_from_edge_dict for weighted edges given as sequence --
    >>> edges_weighted_sequence = (
    ...     ((1, 1), (2, 2)),
    ...     ((2, 5), (3, 5)),
    ...     ((3, 2), (4, 1), (5, 5)),
    ...     ((5, 2), (6, 5)),
    ...     ((5, 5), (1, 1)),
    ...     ((6, 2),)
    ...  )

    >>> next_vertices = nog.adapt_edge_index(edges_weighted_sequence,
    ...     add_inverted=False, labeled=True)
    >>> tuple(edges(next_vertices, range(len(edges_weighted_sequence))).values()
    ...      ) == edges_weighted_sequence
    True

    >>> next_vertices = nog.adapt_edge_index(
    ...    edges_weighted_sequence,add_inverted=True,labeled=True)
    >>> edges(next_vertices, range(len(edges_weighted_sequence)))  # doctest: +NORMALIZE_WHITESPACE
    {0: [(1, 1), (2, 2)], 1: [(0, 1), (2, 5), (3, 5), (4, 1)],
     2: [(0, 2), (1, 5), (3, 2), (4, 1), (5, 5)], 3: [(1, 5), (2, 2), (5, 2), (6, 5)],
     4: [(2, 1), (5, 5), (1, 1)], 5: [(2, 5), (3, 2), (4, 5), (6, 2)]}


    -- next_from_edge_dict for unweighted edges given as dict --
    >>> edges_unweighted_dict = {
    ...    'A': ('B', 'C'),
    ...    'B': ('C', 'D'),
    ...    'C': ('D', 'E', 'F'),
    ...    'D': ('F', 'G'),
    ...    'E': ('F', 'B'),
    ...    'F': ('G', )
    ... }

    >>> next_vertices = nog.adapt_edge_index(
    ...    edges_unweighted_dict,add_inverted=False,labeled=False)
    >>> edges(next_vertices, edges_weighted_dict.keys()) == edges_unweighted_dict
    True

    >>> next_vertices = nog.adapt_edge_index(
    ...    edges_unweighted_dict,add_inverted=True,labeled=False)
    >>> edges(next_vertices, edges_weighted_dict.keys())  # doctest: +NORMALIZE_WHITESPACE
    {'A': ['B', 'C'], 'B': ['A', 'C', 'D', 'E'], 'C': ['A', 'B', 'D', 'E', 'F'],
     'D': ['B', 'C', 'F', 'G'], 'E': ['C', 'F', 'B'], 'F': ['C', 'D', 'E', 'G']}


    -- next_from_edge_dict for unweighted edges given as sequence --
    >>> edges_unweighted_sequence = (
    ...     (1, 2),
    ...     (2, 3),
    ...     (3, 4, 5),
    ...     (5, 6),
    ...     (5, 1),
    ...     (6,)
    ... )

    >>> next_vertices = nog.adapt_edge_index(
    ...    edges_unweighted_sequence,add_inverted=False,labeled=False)
    >>> tuple(edges(next_vertices, range(len(edges_weighted_sequence))).values()) \
        == edges_unweighted_sequence
    True

    >>> next_vertices = nog.adapt_edge_index(
    ...    edges_unweighted_sequence,add_inverted=True,labeled=False)
    >>> edges(next_vertices, range(len(edges_weighted_sequence)))  # doctest: +NORMALIZE_WHITESPACE
    {0: [1, 2], 1: [0, 2, 3, 4], 2: [0, 1, 3, 4, 5], 3: [1, 2, 5, 6],
     4: [2, 5, 1], 5: [2, 3, 4, 6]}
    """


# --------- Tests for all traversals -----------


class PathHandling:
    # noinspection PyShadowingNames
    """-- Error handling in Paths --
     Path detects illegal calls with None as vertex and with non-existing vertices.
     Iteration works for both types of paths both with vertices and with edges.

    -- Unlabeled paths --

    >>> path_unlabeled = nog.PathsOfUnlabeledEdges(dict(), None)
    >>> path_unlabeled[None]  # Calls p.iter_vertices_to_start
    Traceback (most recent call last):
    RuntimeError: Paths: None instead of vertex given.
    >>> tuple(path_unlabeled.iter_vertices_to_start(None))
    Traceback (most recent call last):
    RuntimeError: Paths: None instead of vertex given.
    >>> tuple(path_unlabeled.iter_edges_to_start(None))
    Traceback (most recent call last):
    RuntimeError: Paths: None instead of vertex given.

    >>> tuple(path_unlabeled.iter_edges_to_start(2))
    Traceback (most recent call last):
    RuntimeError: Paths: No path for given vertex.

    >>> path_unlabeled.append_edge(0, 1, None)
    >>> path_unlabeled[1]  # Calls p.iter_vertices_from_start and p.iter_vertices_to_start
    (0, 1)
    >>> tuple(path_unlabeled.iter_edges_from_start(1))  # Also calls p.iter_edges_to_start
    ((0, 1),)

    >>> path_unlabeled.append_edge(1, 2, None)
    >>> path_unlabeled[2]
    (0, 1, 2)
    >>> tuple(path_unlabeled.iter_edges_from_start(2))
    ((0, 1), (1, 2))

    -- Labeled paths --

    >>> path_labeled = nog.PathsOfLabeledEdges(dict(), dict(), None)
    >>> path_labeled[None]
    Traceback (most recent call last):
    RuntimeError: Paths: None instead of vertex given.
    >>> tuple(path_labeled.iter_vertices_to_start(None))
    Traceback (most recent call last):
    RuntimeError: Paths: None instead of vertex given.
    >>> tuple(path_labeled.iter_edges_to_start(None))
    Traceback (most recent call last):
    RuntimeError: Paths: None instead of vertex given.

    >>> tuple(path_labeled.iter_edges_to_start(2))
    Traceback (most recent call last):
    RuntimeError: Paths: No path for given vertex.

    >>> path_labeled.append_edge(0, 1, (0, "labeled"))
    >>> path_labeled[1]  # Calls p.iter_edges_from_start and p.iter_edges_to_start
    ((0, 1, 'labeled'),)
    >>> tuple(path_labeled.iter_vertices_from_start(1))  # Also calls p.iter_vertices_to_start
    (0, 1)

    >>> path_labeled.append_edge(1, 2, (0, "labeled"))
    >>> path_labeled[2]
    ((0, 1, 'labeled'), (1, 2, 'labeled'))
    >>> tuple(path_labeled.iter_vertices_from_start(2))
    (0, 1, 2)
    """


class GraphWithoutEdges:
    # noinspection PyShadowingNames
    """-- Graph without edges --
     No vertices are reported. Exception: topological sorting reports it.
     Strategies detect if labeled paths are demanded, but not the generation of paths.

    >>> def no_next(v, _):
    ...     return ()
    >>> def heuristic(v):
    ...    return 0  # Graph has just une vertex and no edges. So, no heuristic needed.

    >>> for strategy in (nog.TraversalBreadthFirst, nog.TraversalDepthFirst,
    ...                  nog.TraversalShortestPaths, nog.TraversalMinimumSpanningTree,
    ...                  nog.TraversalTopologicalSort):
    ...     list(strategy(no_next).start_from(0))
    []
    []
    []
    []
    [0]
    >>> list(nog.TraversalAStar(no_next).start_from(heuristic, 0))
    []

    >>> traversal = nog.TraversalBreadthFirst(no_next)
    >>> traversal.start_from(0).go_to(1, fail_silently=True) is None
    True
    >>> traversal.start_from(0).go_to(1)
    Traceback (most recent call last):
    KeyError: 'Vertex not found, graph exhausted.'
    >>> list(traversal.start_from(0).go_for_vertices_in((1,), fail_silently=True))
    []
    >>> list(traversal.start_from(0).go_for_vertices_in((1,)))
    Traceback (most recent call last):
    KeyError: 'Not all of the given vertices have been found'


    >>> traversal = nog.TraversalBreadthFirst(no_next)
    >>> traversal = traversal.start_from(0, build_paths=False, labeled_paths=True)
    Traceback (most recent call last):
    RuntimeError: Option labeled_paths without option build_paths.
    >>> traversal = traversal.start_from(start_vertex=0, start_vertices=(0,))
    Traceback (most recent call last):
    RuntimeError: Both start_vertex and start_vertices provided.
    >>> traversal = traversal.start_from()
    Traceback (most recent call last):
    RuntimeError: Neither start_vertex and start_vertices provided.
    >>> traversal = nog.TraversalBreadthFirst()
    Traceback (most recent call last):
    RuntimeError: Neither next_vertices nor next_edges provided.
    >>> traversal = nog.TraversalBreadthFirst(next_edges=no_next, next_vertices=no_next)
    Traceback (most recent call last):
    RuntimeError: Both next_vertices and next_edges provided.
    """


class GraphWithOneEdgeAndPathVariants:
    # noinspection PyShadowingNames
    """-- Graph with one edge --
    Start vertex not reported (exception: topological sorting). First edge followed.
    Paths not build if not demanded, and build if demanded.
    Labeled paths not allowed for unlabeled edges, and build for labeled edges, if demanded.
    Calculation limit raises Exception at exactly correct number of visited vertices.

    >>> def test(traversal, with_is_labeled):
    ...     print(list(traversal.start_from(0)))
    ...     print(traversal.paths is None)
    ...     print(next(traversal.start_from(0)))
    ...     _ = traversal.start_from(0, build_paths=True).go_to(1)
    ...     print(traversal.paths[1])
    ...     if with_is_labeled:
    ...        _ = traversal.start_from(0, build_paths=True, labeled_paths=True).go_to(1)
    ...        print(traversal.paths[1])

    # -- Unlabeled graph --
    >>> def next_vertices(v, _):
    ...     return (1,) if v == 0 else ()

    >>> traversal = nog.TraversalBreadthFirst(next_vertices)
    >>> test(traversal, with_is_labeled=False)
    [1]
    True
    1
    (0, 1)
    >>> traversal = nog.TraversalDepthFirst(next_vertices)
    >>> test(traversal, with_is_labeled=False)
    [1]
    True
    1
    (0, 1)
    >>> traversal = nog.TraversalTopologicalSort(next_vertices)
    >>> test(traversal, with_is_labeled=False)
    [1, 0]
    True
    1
    (0, 1)
    >>> traversal = nog.TraversalTopologicalSort(next_vertices, is_tree=True)
    >>> test(traversal, with_is_labeled=False)
    [1, 0]
    True
    1
    (0, 1)

    >>> nog.TraversalBreadthFirst(next_vertices).start_from(0, build_paths=True, labeled_paths=True)
    Traceback (most recent call last):
    RuntimeError: A labeled path can only be computed from labeled edges.
    >>> nog.TraversalDepthFirst(next_vertices).start_from(0, build_paths=True, labeled_paths=True)
    Traceback (most recent call last):
    RuntimeError: A labeled path can only be computed from labeled edges.
    >>> nog.TraversalTopologicalSort(next_vertices).start_from(
    ...     0, build_paths=True, labeled_paths=True)
    Traceback (most recent call last):
    RuntimeError: A labeled path can only be computed from labeled edges.


    # -- Labeled graph --
    >>> def next_edges(v, _):
    ...     return ((1, 1),) if v == 0 else ()

    >>> traversal = nog.TraversalBreadthFirst(next_edges=next_edges)
    >>> test(traversal, with_is_labeled=True)
    [1]
    True
    1
    (0, 1)
    ((0, 1, 1),)
    >>> traversal = nog.TraversalDepthFirst(next_edges=next_edges)
    >>> test(traversal, with_is_labeled=True)
    [1]
    True
    1
    (0, 1)
    ((0, 1, 1),)
    >>> traversal = nog.TraversalShortestPaths(next_edges=next_edges)
    >>> test(traversal, with_is_labeled=True)
    [1]
    True
    1
    (0, 1)
    ((0, 1, 1),)
    >>> traversal = nog.TraversalMinimumSpanningTree(next_edges=next_edges)
    >>> test(traversal, with_is_labeled=True)
    [1]
    True
    1
    (0, 1)
    ((0, 1, 1),)
    >>> traversal = nog.TraversalTopologicalSort(next_edges=next_edges)
    >>> test(traversal, with_is_labeled=True)
    [1, 0]
    True
    1
    (0, 1)
    ((0, 1, 1),)
    >>> traversal = nog.TraversalTopologicalSort(next_edges=next_edges, is_tree=True)
    >>> test(traversal, with_is_labeled=True)
    [1, 0]
    True
    1
    (0, 1)
    ((0, 1, 1),)


    >>> traversal = nog.TraversalBreadthFirst(next_edges=next_edges)
    >>> _ = list(traversal.start_from(0, calculation_limit=2))
    >>> _ = list(traversal.start_from(0, calculation_limit=1))
    Traceback (most recent call last):
    RuntimeError: Number of visited vertices reached limit
    >>> traversal = nog.TraversalDepthFirst(next_edges=next_edges)
    >>> _ = list(traversal.start_from(0, calculation_limit=2))
    >>> _ = list(traversal.start_from(0, calculation_limit=1))
    Traceback (most recent call last):
    RuntimeError: Number of visited vertices reached limit
    >>> traversal = nog.TraversalTopologicalSort(next_edges=next_edges)
    >>> _ = list(traversal.start_from(0, calculation_limit=2))
    >>> _ = list(traversal.start_from(0, calculation_limit=1))
    Traceback (most recent call last):
    RuntimeError: Number of visited vertices reached limit
    >>> traversal = nog.TraversalTopologicalSort(next_edges=next_edges, is_tree=True)
    >>> _ = list(traversal.start_from(0, calculation_limit=2))
    >>> _ = list(traversal.start_from(0, calculation_limit=1))
    Traceback (most recent call last):
    RuntimeError: Number of visited vertices reached limit
    >>> traversal = nog.TraversalShortestPaths(next_edges=next_edges)
    >>> _ = list(traversal.start_from(0, calculation_limit=2))
    >>> _ = list(traversal.start_from(0, calculation_limit=1))
    Traceback (most recent call last):
    RuntimeError: Number of visited vertices reached limit
    >>> traversal = nog.TraversalMinimumSpanningTree(next_edges=next_edges)
    >>> _ = list(traversal.start_from(0, calculation_limit=2))
    >>> _ = list(traversal.start_from(0, calculation_limit=1))
    Traceback (most recent call last):
    RuntimeError: Number of visited vertices reached limit
    >>> # Test early exceeded limit, during traversal of edges from start vertex in MST
    >>> _ = list(traversal.start_from(0, calculation_limit=0))
    Traceback (most recent call last):
    RuntimeError: Number of visited vertices reached limit


    >>> def heuristic(v):
    ...    return 1 if v == 0 else 0  # heuristic with perfect estimation
    >>> traversal = nog.TraversalAStar(next_edges=next_edges)
    >>> list(traversal.start_from(heuristic, 0))
    [1]
    >>> traversal.paths is None
    True
    >>> _ = traversal.start_from(heuristic, 0, build_paths=True).go_to(1)
    >>> traversal.paths[1]
    (0, 1)
    >>> _ = traversal.start_from(heuristic, 0, build_paths=True, labeled_paths=True).go_to(1)
    >>> traversal.paths[1]
    ((0, 1, 1),)

    >>> traversal = nog.TraversalAStar(next_edges=next_edges)
    >>> _ = list(traversal.start_from(heuristic, 0, calculation_limit=2))
    >>> _ = list(traversal.start_from(heuristic, 0, calculation_limit=1))
    Traceback (most recent call last):
    RuntimeError: Number of visited vertices reached limit
    """


class GraphWithOneEdgeAndVertexToId:
    # noinspection PyShadowingNames
    """-- Graph with one edge, parameter vertex_to_id used --
    For the test, vertices are given as list [int] to have something that is not hashable.
    Start vertex not reported (exception: topological sorting). First edge followed.
    Paths not build if not demanded, and build if demanded.
    Labeled paths not allowed for unlabeled edges, and build for labeled edges, if demanded.
    Calculation limit raises Exception at exactly correct number of visited vertices.

    >>> def next_vertices(v, _):
    ...     return ([1],) if v == [0] else ()
    >>> for strategy in (nog.TraversalBreadthFirst, nog.TraversalDepthFirst,
    ...                  nog.TraversalTopologicalSort):
    ...     print(strategy.__name__)
    ...     traversal = strategy(next_vertices, vertex_to_id=lambda l: l[0])
    ...     list(traversal.start_from([0]))
    ...     traversal.paths is None
    ...     _ = traversal.start_from([0], build_paths=True).go_to([1])
    ...     traversal.paths[[1]]
    TraversalBreadthFirst
    [[1]]
    True
    ([0], [1])
    TraversalDepthFirst
    [[1]]
    True
    ([0], [1])
    TraversalTopologicalSort
    [[1], [0]]
    True
    ([0], [1])


    >>> def next_edges(v, _):
    ...     return (([1], 1),) if v == [0] else ()
    >>> for strategy in (nog.TraversalBreadthFirst, nog.TraversalDepthFirst,
    ...                  nog.TraversalShortestPaths, nog.TraversalMinimumSpanningTree,
    ...                  nog.TraversalTopologicalSort):
    ...     print(strategy.__name__)
    ...     traversal = strategy(next_edges=next_edges, vertex_to_id=lambda l: l[0])
    ...     list(traversal.start_from([0]))
    ...     traversal.paths is None
    ...     _ = traversal.start_from([0], build_paths=True).go_to([1])
    ...     traversal.paths[[1]]
    ...     _ = traversal.start_from([0], build_paths=True, labeled_paths=True).go_to([1])
    ...     traversal.paths[[1]]
    TraversalBreadthFirst
    [[1]]
    True
    ([0], [1])
    (([0], [1], 1),)
    TraversalDepthFirst
    [[1]]
    True
    ([0], [1])
    (([0], [1], 1),)
    TraversalShortestPaths
    [[1]]
    True
    ([0], [1])
    (([0], [1], 1),)
    TraversalMinimumSpanningTree
    [[1]]
    True
    ([0], [1])
    (([0], [1], 1),)
    TraversalTopologicalSort
    [[1], [0]]
    True
    ([0], [1])
    (([0], [1], 1),)


    >>> def heuristic(v):
    ...    return 1 if v == [0] else 0  # heuristic with perfect estimation
    >>> traversal = nog.TraversalAStar(next_edges=next_edges, vertex_to_id=lambda l: l[0])
    >>> list(traversal.start_from(heuristic, [0]))
    [[1]]
    >>> traversal.paths is None
    True
    >>> _ = traversal.start_from(heuristic, [0], build_paths=True).go_to([1])
    >>> traversal.paths[[1]]
    ([0], [1])
    >>> _ = traversal.start_from(heuristic, [0], build_paths=True, labeled_paths=True).go_to([1])
    >>> traversal.paths[[1]]
    (([0], [1], 1),)
    """


class NormalGraphTraversalsWithOrWithoutLabels:
    # noinspection PyShadowingNames
    """-- Small example graph, 3 Traversal strategies that can work with and wihout labels --
    Correct traversal of TraversalBreadthFirst, TraversalDepthFirst, TraversalTopologicalSort.
    For: graph without labels and graph with labeles.
    Without and with vertices defined as already visited (only tested for graph without labels).
    (Uses implementation descisions:
    - TraversalBreadthFirst traverses edges in given order
    - TraversalDepthFirst and TraversalTopologicalSort traverse edges in reversed order)

    1. Unlabeled edges

    >>> def next_vertices(vertex, traversal):
    ...     print(f"Next called: {vertex=} {traversal.depth=} {traversal.visited=}")
    ...     return (1, 2) if vertex == 0 else ((3,) if vertex in (1, 2) else ())

    >>> def this_test(strategy):
    ...     traversal = strategy(next_vertices).start_from(0, build_paths=True)
    ...     for vertex in traversal:
    ...         print(f"Reported: {vertex=} {traversal.depth=} {traversal.visited=}")
    ...         print(f"  traversal.paths[vertex]={traversal.paths[vertex]}")
    ...     print("All paths:", [traversal.paths[vertex] for vertex in range(4)])
    ...     print()
    ...     already_visited = {1}  # vertex 1 already visited
    ...     traversal = strategy(next_vertices).start_from(0, build_paths=True,
    ...         already_visited=already_visited)
    ...     for vertex in traversal:
    ...         print(f"Reported: {vertex=} {traversal.depth=}")
    ...         print(f"  traversal.paths[vertex]={traversal.paths[vertex]}")
    ...     print("All paths:", [traversal.paths[vertex] for vertex in range(4) if vertex!=1])
    ...     print("Already visited:", already_visited)
    ...     return traversal

    >>> traversal = this_test(nog.TraversalBreadthFirst)
    Next called: vertex=0 traversal.depth=0 traversal.visited={0}
    Reported: vertex=1 traversal.depth=1 traversal.visited={0, 1}
      traversal.paths[vertex]=(0, 1)
    Reported: vertex=2 traversal.depth=1 traversal.visited={0, 1, 2}
      traversal.paths[vertex]=(0, 2)
    Next called: vertex=1 traversal.depth=1 traversal.visited={0, 1, 2}
    Reported: vertex=3 traversal.depth=2 traversal.visited={0, 1, 2, 3}
      traversal.paths[vertex]=(0, 1, 3)
    Next called: vertex=2 traversal.depth=1 traversal.visited={0, 1, 2, 3}
    Next called: vertex=3 traversal.depth=2 traversal.visited={0, 1, 2, 3}
    All paths: [(0,), (0, 1), (0, 2), (0, 1, 3)]
    <BLANKLINE>
    Next called: vertex=0 traversal.depth=0 traversal.visited={0, 1}
    Reported: vertex=2 traversal.depth=1
      traversal.paths[vertex]=(0, 2)
    Next called: vertex=2 traversal.depth=1 traversal.visited={0, 1, 2}
    Reported: vertex=3 traversal.depth=2
      traversal.paths[vertex]=(0, 2, 3)
    Next called: vertex=3 traversal.depth=2 traversal.visited={0, 1, 2, 3}
    All paths: [(0,), (0, 2), (0, 2, 3)]
    Already visited: {0, 1, 2, 3}
    >>> traversal.paths[1]
    Traceback (most recent call last):
    RuntimeError: Paths: No path for given vertex.

    >>> traversal = this_test(nog.TraversalDepthFirst)
    Next called: vertex=0 traversal.depth=0 traversal.visited={0}
    Reported: vertex=2 traversal.depth=1 traversal.visited={0, 2}
      traversal.paths[vertex]=(0, 2)
    Next called: vertex=2 traversal.depth=1 traversal.visited={0, 2}
    Reported: vertex=3 traversal.depth=2 traversal.visited={0, 2, 3}
      traversal.paths[vertex]=(0, 2, 3)
    Next called: vertex=3 traversal.depth=2 traversal.visited={0, 2, 3}
    Reported: vertex=1 traversal.depth=1 traversal.visited={0, 1, 2, 3}
      traversal.paths[vertex]=(0, 1)
    Next called: vertex=1 traversal.depth=1 traversal.visited={0, 1, 2, 3}
    All paths: [(0,), (0, 1), (0, 2), (0, 2, 3)]
    <BLANKLINE>
    Next called: vertex=0 traversal.depth=0 traversal.visited={0, 1}
    Reported: vertex=2 traversal.depth=1
      traversal.paths[vertex]=(0, 2)
    Next called: vertex=2 traversal.depth=1 traversal.visited={0, 1, 2}
    Reported: vertex=3 traversal.depth=2
      traversal.paths[vertex]=(0, 2, 3)
    Next called: vertex=3 traversal.depth=2 traversal.visited={0, 1, 2, 3}
    All paths: [(0,), (0, 2), (0, 2, 3)]
    Already visited: {0, 1, 2, 3}
    >>> traversal.paths[1]
    Traceback (most recent call last):
    RuntimeError: Paths: No path for given vertex.

    >>> traversal = this_test(nog.TraversalTopologicalSort)
    Next called: vertex=0 traversal.depth=0 traversal.visited={0}
    Next called: vertex=2 traversal.depth=1 traversal.visited={0, 2}
    Next called: vertex=3 traversal.depth=2 traversal.visited={0, 2, 3}
    Reported: vertex=3 traversal.depth=2 traversal.visited={0, 2, 3}
      traversal.paths[vertex]=(0, 2, 3)
    Reported: vertex=2 traversal.depth=1 traversal.visited={0, 2, 3}
      traversal.paths[vertex]=(0, 2)
    Next called: vertex=1 traversal.depth=1 traversal.visited={0, 1, 2, 3}
    Reported: vertex=1 traversal.depth=1 traversal.visited={0, 1, 2, 3}
      traversal.paths[vertex]=(0, 1)
    Reported: vertex=0 traversal.depth=0 traversal.visited={0, 1, 2, 3}
      traversal.paths[vertex]=(0,)
    All paths: [(0,), (0, 1), (0, 2), (0, 2, 3)]
    <BLANKLINE>
    Next called: vertex=0 traversal.depth=0 traversal.visited={0, 1}
    Next called: vertex=2 traversal.depth=1 traversal.visited={0, 1, 2}
    Next called: vertex=3 traversal.depth=2 traversal.visited={0, 1, 2, 3}
    Reported: vertex=3 traversal.depth=2
      traversal.paths[vertex]=(0, 2, 3)
    Reported: vertex=2 traversal.depth=1
      traversal.paths[vertex]=(0, 2)
    Reported: vertex=0 traversal.depth=0
      traversal.paths[vertex]=(0,)
    All paths: [(0,), (0, 2), (0, 2, 3)]
    Already visited: {0, 1, 2, 3}
    >>> traversal.paths[1]
    Traceback (most recent call last):
    RuntimeError: Paths: No path for given vertex.


    2. Labeled edges

    >>> def next_edges(vertex, traversal):
    ...     print(f"Next called: {vertex=} {traversal.depth=} {traversal.visited=}")
    ...     return ((1, 2), (2, 1)) if vertex == 0 else (((3, 2),) if vertex in (1, 2) else ())

    >>> def this_test(strategy):
    ...     traversal = strategy(next_edges=next_edges).start_from(0, build_paths=True)
    ...     for vertex in traversal:
    ...         print(f"Reported: {vertex=} {traversal.depth=} {traversal.visited=}")
    ...         print(f"  traversal.paths[vertex]={traversal.paths[vertex]}")
    ...     print("All paths:", [traversal.paths[vertex] for vertex in range(4)])

    >>> this_test(nog.TraversalBreadthFirst)
    Next called: vertex=0 traversal.depth=0 traversal.visited={0}
    Reported: vertex=1 traversal.depth=1 traversal.visited={0, 1}
      traversal.paths[vertex]=(0, 1)
    Reported: vertex=2 traversal.depth=1 traversal.visited={0, 1, 2}
      traversal.paths[vertex]=(0, 2)
    Next called: vertex=1 traversal.depth=1 traversal.visited={0, 1, 2}
    Reported: vertex=3 traversal.depth=2 traversal.visited={0, 1, 2, 3}
      traversal.paths[vertex]=(0, 1, 3)
    Next called: vertex=2 traversal.depth=1 traversal.visited={0, 1, 2, 3}
    Next called: vertex=3 traversal.depth=2 traversal.visited={0, 1, 2, 3}
    All paths: [(0,), (0, 1), (0, 2), (0, 1, 3)]

    >>> this_test(nog.TraversalDepthFirst)
    Next called: vertex=0 traversal.depth=0 traversal.visited={0}
    Reported: vertex=2 traversal.depth=1 traversal.visited={0, 2}
      traversal.paths[vertex]=(0, 2)
    Next called: vertex=2 traversal.depth=1 traversal.visited={0, 2}
    Reported: vertex=3 traversal.depth=2 traversal.visited={0, 2, 3}
      traversal.paths[vertex]=(0, 2, 3)
    Next called: vertex=3 traversal.depth=2 traversal.visited={0, 2, 3}
    Reported: vertex=1 traversal.depth=1 traversal.visited={0, 1, 2, 3}
      traversal.paths[vertex]=(0, 1)
    Next called: vertex=1 traversal.depth=1 traversal.visited={0, 1, 2, 3}
    All paths: [(0,), (0, 1), (0, 2), (0, 2, 3)]

    >>> this_test(nog.TraversalTopologicalSort)
    Next called: vertex=0 traversal.depth=0 traversal.visited={0}
    Next called: vertex=2 traversal.depth=1 traversal.visited={0, 2}
    Next called: vertex=3 traversal.depth=2 traversal.visited={0, 2, 3}
    Reported: vertex=3 traversal.depth=2 traversal.visited={0, 2, 3}
      traversal.paths[vertex]=(0, 2, 3)
    Reported: vertex=2 traversal.depth=1 traversal.visited={0, 2, 3}
      traversal.paths[vertex]=(0, 2)
    Next called: vertex=1 traversal.depth=1 traversal.visited={0, 1, 2, 3}
    Reported: vertex=1 traversal.depth=1 traversal.visited={0, 1, 2, 3}
      traversal.paths[vertex]=(0, 1)
    Reported: vertex=0 traversal.depth=0 traversal.visited={0, 1, 2, 3}
      traversal.paths[vertex]=(0,)
    All paths: [(0,), (0, 1), (0, 2), (0, 2, 3)]


    3. is_tree

    >>> def next_vertices(vertex, traversal):
    ...     print(f"Next called: {vertex=} {traversal.depth=} {traversal.visited=}")
    ...     return (2*vertex, 2*vertex+1) if 1 <= vertex <= 3 else ()

    >>> def this_test(strategy):
    ...     traversal = strategy(next_vertices, is_tree=True).start_from(1, build_paths=True)
    ...     for vertex in traversal:
    ...         print(f"Reported: {vertex=} {traversal.depth=} {traversal.visited=}")
    ...         print(f"  traversal.paths[vertex]={traversal.paths[vertex]}")
    ...     print("All paths:", [traversal.paths[vertex] for vertex in range(1,8)])

    >>> this_test(nog.TraversalBreadthFirst)
    Next called: vertex=1 traversal.depth=0 traversal.visited=set()
    Reported: vertex=2 traversal.depth=1 traversal.visited=set()
      traversal.paths[vertex]=(1, 2)
    Reported: vertex=3 traversal.depth=1 traversal.visited=set()
      traversal.paths[vertex]=(1, 3)
    Next called: vertex=2 traversal.depth=1 traversal.visited=set()
    Reported: vertex=4 traversal.depth=2 traversal.visited=set()
      traversal.paths[vertex]=(1, 2, 4)
    Reported: vertex=5 traversal.depth=2 traversal.visited=set()
      traversal.paths[vertex]=(1, 2, 5)
    Next called: vertex=3 traversal.depth=1 traversal.visited=set()
    Reported: vertex=6 traversal.depth=2 traversal.visited=set()
      traversal.paths[vertex]=(1, 3, 6)
    Reported: vertex=7 traversal.depth=2 traversal.visited=set()
      traversal.paths[vertex]=(1, 3, 7)
    Next called: vertex=4 traversal.depth=2 traversal.visited=set()
    Next called: vertex=5 traversal.depth=2 traversal.visited=set()
    Next called: vertex=6 traversal.depth=2 traversal.visited=set()
    Next called: vertex=7 traversal.depth=2 traversal.visited=set()
    All paths: [(1,), (1, 2), (1, 3), (1, 2, 4), (1, 2, 5), (1, 3, 6), (1, 3, 7)]

    >>> this_test(nog.TraversalDepthFirst)
    Next called: vertex=1 traversal.depth=0 traversal.visited=set()
    Reported: vertex=3 traversal.depth=1 traversal.visited=set()
      traversal.paths[vertex]=(1, 3)
    Next called: vertex=3 traversal.depth=1 traversal.visited=set()
    Reported: vertex=7 traversal.depth=2 traversal.visited=set()
      traversal.paths[vertex]=(1, 3, 7)
    Next called: vertex=7 traversal.depth=2 traversal.visited=set()
    Reported: vertex=6 traversal.depth=3 traversal.visited=set()
      traversal.paths[vertex]=(1, 3, 6)
    Next called: vertex=6 traversal.depth=3 traversal.visited=set()
    Reported: vertex=2 traversal.depth=4 traversal.visited=set()
      traversal.paths[vertex]=(1, 2)
    Next called: vertex=2 traversal.depth=4 traversal.visited=set()
    Reported: vertex=5 traversal.depth=5 traversal.visited=set()
      traversal.paths[vertex]=(1, 2, 5)
    Next called: vertex=5 traversal.depth=5 traversal.visited=set()
    Reported: vertex=4 traversal.depth=6 traversal.visited=set()
      traversal.paths[vertex]=(1, 2, 4)
    Next called: vertex=4 traversal.depth=6 traversal.visited=set()
    All paths: [(1,), (1, 2), (1, 3), (1, 2, 4), (1, 2, 5), (1, 3, 6), (1, 3, 7)]

    >>> this_test(nog.TraversalTopologicalSort)
    Next called: vertex=1 traversal.depth=0 traversal.visited=set()
    Next called: vertex=3 traversal.depth=1 traversal.visited=set()
    Next called: vertex=7 traversal.depth=2 traversal.visited=set()
    Reported: vertex=7 traversal.depth=2 traversal.visited=set()
      traversal.paths[vertex]=(1, 3, 7)
    Next called: vertex=6 traversal.depth=2 traversal.visited=set()
    Reported: vertex=6 traversal.depth=2 traversal.visited=set()
      traversal.paths[vertex]=(1, 3, 6)
    Reported: vertex=3 traversal.depth=1 traversal.visited=set()
      traversal.paths[vertex]=(1, 3)
    Next called: vertex=2 traversal.depth=1 traversal.visited=set()
    Next called: vertex=5 traversal.depth=2 traversal.visited=set()
    Reported: vertex=5 traversal.depth=2 traversal.visited=set()
      traversal.paths[vertex]=(1, 2, 5)
    Next called: vertex=4 traversal.depth=2 traversal.visited=set()
    Reported: vertex=4 traversal.depth=2 traversal.visited=set()
      traversal.paths[vertex]=(1, 2, 4)
    Reported: vertex=2 traversal.depth=1 traversal.visited=set()
      traversal.paths[vertex]=(1, 2)
    Reported: vertex=1 traversal.depth=0 traversal.visited=set()
      traversal.paths[vertex]=(1,)
    All paths: [(1,), (1, 2), (1, 3), (1, 2, 4), (1, 2, 5), (1, 3, 6), (1, 3, 7)]
    """


class NormalGraphTraversalsWithLabels:
    # noinspection PyShadowingNames
    """-- Small example graph. The three strategies that need edge weights. --
    For TraversalShortestPaths, test option known_distances, too.

    >>> def next_edges(vertex, traversal):
    ...     print(f"Next called: {vertex=} {traversal.distance=} {traversal.depth=}")
    ...     return ((1, 2), (2, 1)) if vertex == 0 else (((3, 2),) if vertex in (1, 2) else ())

    >>> traversal = nog.TraversalShortestPaths(next_edges=next_edges)
    >>> traversal = traversal.start_from(0, build_paths=True)
    >>> reported = []
    >>> for vertex in traversal:
    ...     reported.append(vertex)
    ...     print(f"Reported: {vertex=} {traversal.distance=} {traversal.depth=}")
    ...     print("distances for reported vertices: " +
    ...           f"{str({vertex: traversal.distances[vertex] for vertex in reported})} " +
    ...           f"traversal.paths={traversal.paths[vertex]}")
    Next called: vertex=0 traversal.distance=0 traversal.depth=0
    Reported: vertex=2 traversal.distance=1 traversal.depth=1
    distances for reported vertices: {2: 0} traversal.paths=(0, 2)
    Next called: vertex=2 traversal.distance=1 traversal.depth=1
    Reported: vertex=1 traversal.distance=2 traversal.depth=1
    distances for reported vertices: {2: 0, 1: 0} traversal.paths=(0, 1)
    Next called: vertex=1 traversal.distance=2 traversal.depth=1
    Reported: vertex=3 traversal.distance=3 traversal.depth=2
    distances for reported vertices: {2: 0, 1: 0, 3: 0} traversal.paths=(0, 2, 3)
    Next called: vertex=3 traversal.distance=3 traversal.depth=2
    >>> print("All paths:", [traversal.paths[vertex] for vertex in range(4)])
    All paths: [(0,), (0, 1), (0, 2), (0, 2, 3)]

    Start vertex starts with distance 2, and we pretend to have a path to 1 with distance 0.
    Effect: All reported distances are 2 higher than in the test before, because the traversal
    starts at distance 2, and vertex 1 is not visited and reported at all, because no path
    can be found that beats this "already found" low distance.
    >>> infinity = float('infinity')
    >>> known_distances = collections.defaultdict(lambda: infinity, ((0, 2), (1, 0)))
    >>> traversal = traversal.start_from(0, build_paths=True, known_distances=known_distances)
    >>> for vertex in traversal:
    ...     print(f"Reported: {vertex=} {traversal.distance=} {traversal.depth=} "
    ...         + f"traversal.paths={traversal.paths[vertex]}")
    Next called: vertex=0 traversal.distance=2 traversal.depth=0
    Reported: vertex=2 traversal.distance=3 traversal.depth=1 traversal.paths=(0, 2)
    Next called: vertex=2 traversal.distance=3 traversal.depth=1
    Reported: vertex=3 traversal.distance=5 traversal.depth=2 traversal.paths=(0, 2, 3)
    Next called: vertex=3 traversal.distance=5 traversal.depth=2
    >>> print("All paths:", [traversal.paths[vertex] for vertex in range(4) if vertex != 1])
    All paths: [(0,), (0, 2), (0, 2, 3)]
    >>> print("Known distances:", dict(known_distances))
    Known distances: {0: 0, 1: 0, 2: 0, 3: 0}


    >>> def next_edges(vertex, traversal):
    ...     print(f"Next called: {vertex=} {traversal.edge=}")
    ...     return ((1, 2), (2, 1)) if vertex == 0 else (((3, 3),) if vertex in (1, 2) else ())
    >>> traversal = nog.TraversalMinimumSpanningTree(next_edges=next_edges)
    >>> traversal = traversal.start_from(0, build_paths=True)
    >>> for vertex in traversal:
    ...     print(f"Reported: {vertex=} {traversal.edge=}"
    ...         + f"traversal.paths={traversal.paths[vertex]}")
    Next called: vertex=0 traversal.edge=None
    Reported: vertex=2 traversal.edge=(0, 2, 1)traversal.paths=(0, 2)
    Next called: vertex=2 traversal.edge=(0, 2, 1)
    Reported: vertex=1 traversal.edge=(0, 1, 2)traversal.paths=(0, 1)
    Next called: vertex=1 traversal.edge=(0, 1, 2)
    Reported: vertex=3 traversal.edge=(2, 3, 3)traversal.paths=(0, 2, 3)
    Next called: vertex=3 traversal.edge=(2, 3, 3)
    >>> print("All paths:", [traversal.paths[vertex] for vertex in range(4)])
    All paths: [(0,), (0, 1), (0, 2), (0, 2, 3)]


    Test TraversaAStar. Typically, one would use go_to(3) for our goal vertex 3, but for
    the test purposes we continue to iterate the rest of the graph.
    >>> def heuristic(v):
    ...    return {0:6, 1:1, 2:2, 3:0, 4:infinity}[v]  # makes 2nd best path visited first
    >>> def next_edges(vertex, traversal):
    ...     print(f"Next called: {vertex=} {traversal.path_length=} {traversal.depth=}")
    ...     return {0: ((1, 3), (2, 3), (4, 1)), 1: ((3, 3),), 2: ((3, 2),), 3:(), 4:()}[vertex]
    >>> traversal = nog.TraversalAStar(next_edges=next_edges)
    >>> traversal = traversal.start_from(heuristic, 0, build_paths=True)
    >>> for vertex in traversal:
    ...     print(f"Reported: {vertex=} {traversal.path_length=} {traversal.depth=} "
    ...         + f"traversal.paths={traversal.paths[vertex]}")
    Next called: vertex=0 traversal.path_length=0 traversal.depth=0
    Reported: vertex=1 traversal.path_length=3 traversal.depth=1 traversal.paths=(0, 1)
    Next called: vertex=1 traversal.path_length=3 traversal.depth=1
    Reported: vertex=2 traversal.path_length=3 traversal.depth=1 traversal.paths=(0, 2)
    Next called: vertex=2 traversal.path_length=3 traversal.depth=1
    Reported: vertex=3 traversal.path_length=5 traversal.depth=2 traversal.paths=(0, 2, 3)
    Next called: vertex=3 traversal.path_length=5 traversal.depth=2
    Reported: vertex=4 traversal.path_length=1 traversal.depth=1 traversal.paths=(0, 4)
    Next called: vertex=4 traversal.path_length=1 traversal.depth=1
    >>> print("All paths:", [traversal.paths[vertex] for vertex in range(5)])
    All paths: [(0,), (0, 1), (0, 2), (0, 2, 3), (0, 4)]

    Variant of the test with option known_distances.
    For the start vertex, we define a start distance. Effect: All distances are now 2 higher.
    For vertex 1, we define an artificial and low distance of 0. Effect: it is not visited.
    >>> known_distances = collections.defaultdict(lambda: infinity, ((0, 2),(1, 0)))
    >>> known_path_length_guesses = collections.defaultdict(lambda: infinity)
    >>> traversal = traversal.start_from(heuristic, 0, build_paths=True,
    ...     known_distances=known_distances, known_path_length_guesses=known_path_length_guesses)
    >>> for vertex in traversal:
    ...     print(f"Reported: {vertex=} {traversal.path_length=} {traversal.depth=} "
    ...         + f"traversal.paths={traversal.paths[vertex]}")
    Next called: vertex=0 traversal.path_length=2 traversal.depth=0
    Reported: vertex=2 traversal.path_length=5 traversal.depth=1 traversal.paths=(0, 2)
    Next called: vertex=2 traversal.path_length=5 traversal.depth=1
    Reported: vertex=3 traversal.path_length=7 traversal.depth=2 traversal.paths=(0, 2, 3)
    Next called: vertex=3 traversal.path_length=7 traversal.depth=2
    Reported: vertex=4 traversal.path_length=3 traversal.depth=1 traversal.paths=(0, 4)
    Next called: vertex=4 traversal.path_length=3 traversal.depth=1
    >>> print("All paths:", [traversal.paths[vertex] for vertex in range(5) if vertex!=1])
    All paths: [(0,), (0, 2), (0, 2, 3), (0, 4)]
    >>> print("Distance at goal:", known_distances[3])
    Distance at goal: 7
    >>> print("Best distances found so far:", dict(known_distances))
    Best distances found so far: {0: 2, 1: 0, 2: 5, 4: 3, 3: 7}
    >>> print("Best path len guesses found so far):", dict(known_path_length_guesses))
    Best path len guesses found so far): {0: 8, 2: 7, 4: inf, 3: 7}


    All algorithms, with option is_tree:

    >>> def next_edges(vertex, traversal):
    ...     print(f"Next called: {vertex=} {traversal.depth=} {traversal.distance=}")
    ...     return ((2*vertex, 2*vertex), (2*vertex+1, 2*vertex+1)) if 1 <= vertex <= 3 else ()
    >>> def this_test(strategy):
    ...     traversal = strategy(next_edges, is_tree=True).start_from(1, build_paths=True)
    ...     for vertex in traversal:
    ...         print(f"Reported: {vertex=} {traversal.depth=} {traversal.distance=} " +
    ...             f"traversal.paths={traversal.paths[vertex]}")
    ...     print("All paths:", [traversal.paths[vertex] for vertex in range(1,8)])
    ...     print("All distances:", dict(traversal.distances))
    >>> this_test(nog.TraversalShortestPaths)
    Next called: vertex=1 traversal.depth=0 traversal.distance=0
    Reported: vertex=2 traversal.depth=1 traversal.distance=2 traversal.paths=(1, 2)
    Next called: vertex=2 traversal.depth=1 traversal.distance=2
    Reported: vertex=3 traversal.depth=1 traversal.distance=3 traversal.paths=(1, 3)
    Next called: vertex=3 traversal.depth=1 traversal.distance=3
    Reported: vertex=4 traversal.depth=2 traversal.distance=6 traversal.paths=(1, 2, 4)
    Next called: vertex=4 traversal.depth=2 traversal.distance=6
    Reported: vertex=5 traversal.depth=2 traversal.distance=7 traversal.paths=(1, 2, 5)
    Next called: vertex=5 traversal.depth=2 traversal.distance=7
    Reported: vertex=6 traversal.depth=2 traversal.distance=9 traversal.paths=(1, 3, 6)
    Next called: vertex=6 traversal.depth=2 traversal.distance=9
    Reported: vertex=7 traversal.depth=2 traversal.distance=10 traversal.paths=(1, 3, 7)
    Next called: vertex=7 traversal.depth=2 traversal.distance=10
    All paths: [(1,), (1, 2), (1, 3), (1, 2, 4), (1, 2, 5), (1, 3, 6), (1, 3, 7)]
    All distances: {1: 0}

    >>> this_test(nog.TraversalMinimumSpanningTree)
    Traceback (most recent call last):
    TypeError: TraversalMinimumSpanningTree.__init__() got an unexpected keyword argument 'is_tree'

    Test TraversaAStar. Typically, one would use go_to(6) for our goal vertex 6, but for
    the test purposes we continue to iterate the rest of the graph.
    >>> def next_edges(vertex, traversal):
    ...     print(f"Next called: {vertex=} {traversal.depth=} {traversal.path_length=}")
    ...     return ((2*vertex, 2*vertex), (2*vertex+1, 2*vertex+1)) if 1 <= vertex <= 3 else ()
    >>> def heuristic(v):
    ...    return {6:0, 3:3}.get(v, 11)
    >>> traversal = nog.TraversalAStar(next_edges=next_edges)
    >>> traversal = traversal.start_from(heuristic, 1, build_paths=True)
    >>> for vertex in traversal:
    ...     print(f"Reported: {vertex=} {traversal.depth=} {traversal.path_length=} "
    ...         + f"traversal.paths={traversal.paths[vertex]}")
    Next called: vertex=1 traversal.depth=0 traversal.path_length=0
    Reported: vertex=3 traversal.depth=1 traversal.path_length=3 traversal.paths=(1, 3)
    Next called: vertex=3 traversal.depth=1 traversal.path_length=3
    Reported: vertex=6 traversal.depth=2 traversal.path_length=9 traversal.paths=(1, 3, 6)
    Next called: vertex=6 traversal.depth=2 traversal.path_length=9
    Reported: vertex=2 traversal.depth=1 traversal.path_length=2 traversal.paths=(1, 2)
    Next called: vertex=2 traversal.depth=1 traversal.path_length=2
    Reported: vertex=4 traversal.depth=2 traversal.path_length=6 traversal.paths=(1, 2, 4)
    Next called: vertex=4 traversal.depth=2 traversal.path_length=6
    Reported: vertex=5 traversal.depth=2 traversal.path_length=7 traversal.paths=(1, 2, 5)
    Next called: vertex=5 traversal.depth=2 traversal.path_length=7
    Reported: vertex=7 traversal.depth=2 traversal.path_length=10 traversal.paths=(1, 3, 7)
    Next called: vertex=7 traversal.depth=2 traversal.path_length=10
    """


class SpecialCases:
    # noinspection PyShadowingNames
    """ TraversalDepthFirst, vertex seams to be not visited at first,
    but becomes visited before it is got from the stack"
    >>> edges = {0: (1, 2), 2: (1,)}
    >>> def next_vertices(vertex, _):
    ...     return edges.get(vertex, ())
    >>> list(nog.TraversalDepthFirst(next_vertices).start_from(0))
    [2, 1]
    """
    # inspection PyShadowingNames


class MultipleStartVerticesTraversalsWithOrWithoutLabels:
    # noinspection PyShadowingNames
    """-- Traversal with multiple start vertices, first 3 traversal strategies --
    Correct traversal of TraversalBreadthFirst, TraversalDepthFirst, TraversalTopologicalSort
    in case of multiple start vertices. No traversal in case of no start vertex.
    (Uses implementation descisions:
    - All strategies travers start vertices in given order
    - TraversalBreadthFirst traverses edges in given order
    - TraversalDepthFirst and TraversalTopologicalSort traverse edges in reversed order)

    1. Unlabeled edges
    >>> def next_vertices(vertex, traversal):
    ...     print(f"Next called: {vertex=} {traversal.depth=}")
    ...     return {0: (2,), 1: (3,), 2: (3, 4), 3: (4,), 4: ()}[vertex]
    >>> for strategy in (nog.TraversalBreadthFirst, nog.TraversalDepthFirst,
    ...                  nog.TraversalTopologicalSort):
    ...     print("- Strategy:", strategy.__name__, "-")
    ...     sv = iter((0,1))  # an iterator (can be iterated only once) should be enough for lib
    ...     traversal = strategy(next_vertices).start_from(start_vertices=sv, build_paths=True)
    ...     for vertex in traversal:
    ...         print(f"Reported: {vertex=} {traversal.depth=} "
    ...             + f"traversal.paths={traversal.paths[vertex]}")
    ...     print("All paths:", [traversal.paths[vertex] for vertex in range(5)])
    ...     traversal = strategy(next_vertices).start_from(start_vertices=(), build_paths=True)
    ...     list(traversal)
    ...     print()
    - Strategy: TraversalBreadthFirst -
    Next called: vertex=0 traversal.depth=0
    Reported: vertex=2 traversal.depth=1 traversal.paths=(0, 2)
    Next called: vertex=1 traversal.depth=0
    Reported: vertex=3 traversal.depth=1 traversal.paths=(1, 3)
    Next called: vertex=2 traversal.depth=1
    Reported: vertex=4 traversal.depth=2 traversal.paths=(0, 2, 4)
    Next called: vertex=3 traversal.depth=1
    Next called: vertex=4 traversal.depth=2
    All paths: [(0,), (1,), (0, 2), (1, 3), (0, 2, 4)]
    []
    <BLANKLINE>
    - Strategy: TraversalDepthFirst -
    Next called: vertex=1 traversal.depth=0
    Reported: vertex=3 traversal.depth=1 traversal.paths=(1, 3)
    Next called: vertex=3 traversal.depth=1
    Reported: vertex=4 traversal.depth=2 traversal.paths=(1, 3, 4)
    Next called: vertex=4 traversal.depth=2
    Next called: vertex=0 traversal.depth=0
    Reported: vertex=2 traversal.depth=1 traversal.paths=(0, 2)
    Next called: vertex=2 traversal.depth=1
    All paths: [(0,), (1,), (0, 2), (1, 3), (1, 3, 4)]
    []
    <BLANKLINE>
    - Strategy: TraversalTopologicalSort -
    Next called: vertex=1 traversal.depth=0
    Next called: vertex=3 traversal.depth=1
    Next called: vertex=4 traversal.depth=2
    Reported: vertex=4 traversal.depth=2 traversal.paths=(1, 3, 4)
    Reported: vertex=3 traversal.depth=1 traversal.paths=(1, 3)
    Reported: vertex=1 traversal.depth=0 traversal.paths=(1,)
    Next called: vertex=0 traversal.depth=0
    Next called: vertex=2 traversal.depth=1
    Reported: vertex=2 traversal.depth=1 traversal.paths=(0, 2)
    Reported: vertex=0 traversal.depth=0 traversal.paths=(0,)
    All paths: [(0,), (1,), (0, 2), (1, 3), (1, 3, 4)]
    []
    <BLANKLINE>

    1. Labeled edges
    >>> def next_edges(vertex, traversal):
    ...     print(f"Next called: {vertex=} {traversal.depth=}")
    ...     return {0: ((2, 1),), 1: ((3, 1),), 2: ((3, 1), (4, 1)), 3: ((4, 1),), 4: ()}[vertex]
    >>> for strategy in (nog.TraversalBreadthFirst, nog.TraversalDepthFirst,
    ...                  nog.TraversalTopologicalSort):
    ...     print("- Strategy:", strategy.__name__, "-")
    ...     traversal = strategy(next_edges=next_edges)
    ...     sv = iter((0,1))  # an iterator (can be iterated only once) should be enough for lib
    ...     traversal = traversal.start_from(start_vertices=sv, build_paths=True)
    ...     for vertex in traversal:
    ...         print(f"Reported: {vertex=} {traversal.depth=} "
    ...             + f"traversal.paths={traversal.paths[vertex]}")
    ...     print("All paths:", [traversal.paths[vertex] for vertex in range(4)])
    ...     traversal = strategy(next_vertices).start_from(start_vertices=(), build_paths=True)
    ...     list(traversal)
    ...     print()
    - Strategy: TraversalBreadthFirst -
    Next called: vertex=0 traversal.depth=0
    Reported: vertex=2 traversal.depth=1 traversal.paths=(0, 2)
    Next called: vertex=1 traversal.depth=0
    Reported: vertex=3 traversal.depth=1 traversal.paths=(1, 3)
    Next called: vertex=2 traversal.depth=1
    Reported: vertex=4 traversal.depth=2 traversal.paths=(0, 2, 4)
    Next called: vertex=3 traversal.depth=1
    Next called: vertex=4 traversal.depth=2
    All paths: [(0,), (1,), (0, 2), (1, 3)]
    []
    <BLANKLINE>
    - Strategy: TraversalDepthFirst -
    Next called: vertex=1 traversal.depth=0
    Reported: vertex=3 traversal.depth=1 traversal.paths=(1, 3)
    Next called: vertex=3 traversal.depth=1
    Reported: vertex=4 traversal.depth=2 traversal.paths=(1, 3, 4)
    Next called: vertex=4 traversal.depth=2
    Next called: vertex=0 traversal.depth=0
    Reported: vertex=2 traversal.depth=1 traversal.paths=(0, 2)
    Next called: vertex=2 traversal.depth=1
    All paths: [(0,), (1,), (0, 2), (1, 3)]
    []
    <BLANKLINE>
    - Strategy: TraversalTopologicalSort -
    Next called: vertex=1 traversal.depth=0
    Next called: vertex=3 traversal.depth=1
    Next called: vertex=4 traversal.depth=2
    Reported: vertex=4 traversal.depth=2 traversal.paths=(1, 3, 4)
    Reported: vertex=3 traversal.depth=1 traversal.paths=(1, 3)
    Reported: vertex=1 traversal.depth=0 traversal.paths=(1,)
    Next called: vertex=0 traversal.depth=0
    Next called: vertex=2 traversal.depth=1
    Reported: vertex=2 traversal.depth=1 traversal.paths=(0, 2)
    Reported: vertex=0 traversal.depth=0 traversal.paths=(0,)
    All paths: [(0,), (1,), (0, 2), (1, 3)]
    []
    <BLANKLINE>
    """


class MultipleStartVerticesTraversalsWithLabels:
    # noinspection PyShadowingNames
    """-- Traversal with multiple start vertex, last 3 traversal strategies --
    Correct traversal in case of multiple start vertices. No traversal in case of no start vertex.

    >>> def next_edges(vertex, traversal):
    ...     print(f"Next called: {vertex=} {traversal.distance=} {traversal.depth=}")
    ...     return {0: ((2, 1),), 1: ((3, 1),), 2: ((3, 1), (4, 1)), 3: ((4, 1),), 4: ()}[vertex]
    >>> traversal = nog.TraversalShortestPaths(next_edges=next_edges)
    >>> sv = iter((0,1))  # an iterator (can be iterated only once) should be enough for lib
    >>> traversal = traversal.start_from(start_vertices=sv, build_paths=True)
    >>> for vertex in traversal:
    ...     print(f"Reported: {vertex=} {traversal.distance=} {traversal.depth=} "
    ...         + f"traversal.paths={traversal.paths[vertex]}")
    Next called: vertex=1 traversal.distance=0 traversal.depth=0
    Next called: vertex=0 traversal.distance=0 traversal.depth=0
    Reported: vertex=2 traversal.distance=1 traversal.depth=1 traversal.paths=(0, 2)
    Next called: vertex=2 traversal.distance=1 traversal.depth=1
    Reported: vertex=3 traversal.distance=1 traversal.depth=1 traversal.paths=(1, 3)
    Next called: vertex=3 traversal.distance=1 traversal.depth=1
    Reported: vertex=4 traversal.distance=2 traversal.depth=2 traversal.paths=(0, 2, 4)
    Next called: vertex=4 traversal.distance=2 traversal.depth=2
    >>> print("All paths:", [traversal.paths[vertex] for vertex in range(5)])
    All paths: [(0,), (1,), (0, 2), (1, 3), (0, 2, 4)]
    >>> traversal = traversal.start_from(start_vertices=(), build_paths=True)
    >>> list(traversal)
    []

    >>> def next_edges(vertex, traversal):
    ...     print(f"Next called: {vertex=} {traversal.edge=}")
    ...     return {0: ((2, 1),), 1: ((3, 1),), 2: ((3, 1), (4, 1)), 3: ((4, 1),), 4: ()}[vertex]
    >>> traversal = nog.TraversalMinimumSpanningTree(next_edges=next_edges)
    >>> sv = iter((0,1))  # an iterator (can be iterated only once) should be enough for lib
    >>> traversal = traversal.start_from(start_vertices=sv, build_paths=True)
    >>> for vertex in traversal:
    ...     print(f"Reported: {vertex=} {traversal.edge=}"
    ...         + f"traversal.paths={traversal.paths[vertex]}")
    Next called: vertex=0 traversal.edge=None
    Next called: vertex=1 traversal.edge=None
    Reported: vertex=2 traversal.edge=(0, 2, 1)traversal.paths=(0, 2)
    Next called: vertex=2 traversal.edge=(0, 2, 1)
    Reported: vertex=3 traversal.edge=(1, 3, 1)traversal.paths=(1, 3)
    Next called: vertex=3 traversal.edge=(1, 3, 1)
    Reported: vertex=4 traversal.edge=(2, 4, 1)traversal.paths=(0, 2, 4)
    Next called: vertex=4 traversal.edge=(2, 4, 1)
    >>> print("All paths:", [traversal.paths[vertex] for vertex in range(5)])
    All paths: [(0,), (1,), (0, 2), (1, 3), (0, 2, 4)]
    >>> traversal = traversal.start_from(start_vertices=(), build_paths=True)
    >>> list(traversal)
    []

    >>> def heuristic(vertex):
    ...    return {0:3, 1:2, 2:2, 3:1, 4:9, 5:0}[vertex] # heuristic with perfect estimation
    >>> def next_edges(vertex, traversal):
    ...     print(f"Next called: {vertex=} {traversal.path_length=} {traversal.depth=}")
    ...     return {0: ((2, 1),), 1: ((3, 1), (4, 1)), 2: ((3, 1),), 3: ((5, 1),), 4: (), 5:()
    ...         }[vertex]
    >>> traversal = nog.TraversalAStar(next_edges=next_edges)
    >>> sv = iter((0,1))  # an iterator (can be iterated only once) should be enough for lib
    >>> traversal = traversal.start_from(heuristic, start_vertices=sv, build_paths=True)
    >>> for vertex in traversal:
    ...     print(f"Reported: {vertex=} {traversal.path_length=} {traversal.depth=} "
    ...         + f"traversal.paths={traversal.paths[vertex]}")
    Next called: vertex=1 traversal.path_length=0 traversal.depth=0
    Reported: vertex=3 traversal.path_length=1 traversal.depth=1 traversal.paths=(1, 3)
    Next called: vertex=3 traversal.path_length=1 traversal.depth=1
    Reported: vertex=5 traversal.path_length=2 traversal.depth=2 traversal.paths=(1, 3, 5)
    Next called: vertex=5 traversal.path_length=2 traversal.depth=2
    Next called: vertex=0 traversal.path_length=0 traversal.depth=0
    Reported: vertex=2 traversal.path_length=1 traversal.depth=1 traversal.paths=(0, 2)
    Next called: vertex=2 traversal.path_length=1 traversal.depth=1
    Reported: vertex=4 traversal.path_length=1 traversal.depth=1 traversal.paths=(1, 4)
    Next called: vertex=4 traversal.path_length=1 traversal.depth=1
    >>> print("All paths:", [traversal.paths[vertex] for vertex in range(6)])
    All paths: [(0,), (1,), (0, 2), (1, 3), (1, 4), (1, 3, 5)]
    >>> traversal = traversal.start_from(heuristic, start_vertices=(), build_paths=True)
    >>> list(traversal)
    []
    """


class InitiationForgotten:
    # noinspection PyTypeChecker
    """Check if the library detects the mistake that start_from or one of the go_... methods are
    called on a traversal class instead of an object, i.e., the round brackets after the class name
    have been forgotten.

    In the following test code, such calls are intentionally made, an since these are typing
    errors, that code inspection is disabled for them in the line on top of the class.

    >>> nog.TraversalBreadthFirst.start_from(None)
    Traceback (most recent call last):
    RuntimeError: Method start_from can only be called on a Traversal object.
    >>> nog.TraversalBreadthFirst.__iter__(None)
    Traceback (most recent call last):
    RuntimeError: Method go can only be called on a Traversal object.
    >>> nog.TraversalBreadthFirst.go_to(None, None)
    Traceback (most recent call last):
    RuntimeError: Method go_to can only be called on a Traversal object.
    >>> nog.TraversalBreadthFirst.go_for_vertices_in(None, None)
    Traceback (most recent call last):
    RuntimeError: Method go_for_vertices_in can only be called on a Traversal object.
    >>> nog.TraversalBreadthFirst.go_for_depth_range(None, None, None)
    Traceback (most recent call last):
    RuntimeError: Method go_for_depth_range can only be called on a Traversal object.

    >>> nog.TraversalDepthFirst.start_from(None)
    Traceback (most recent call last):
    RuntimeError: Method start_from can only be called on a Traversal object.
    >>> nog.TraversalDepthFirst.__iter__(None)
    Traceback (most recent call last):
    RuntimeError: Method go can only be called on a Traversal object.
    >>> nog.TraversalDepthFirst.go_to(None, None)
    Traceback (most recent call last):
    RuntimeError: Method go_to can only be called on a Traversal object.
    >>> nog.TraversalDepthFirst.go_for_vertices_in(None, None)
    Traceback (most recent call last):
    RuntimeError: Method go_for_vertices_in can only be called on a Traversal object.

    >>> nog.TraversalShortestPaths.start_from(None)
    Traceback (most recent call last):
    RuntimeError: Method start_from can only be called on a Traversal object.
    >>> nog.TraversalShortestPaths.__iter__(None)
    Traceback (most recent call last):
    RuntimeError: Method go can only be called on a Traversal object.
    >>> nog.TraversalShortestPaths.go_to(None, None)
    Traceback (most recent call last):
    RuntimeError: Method go_to can only be called on a Traversal object.
    >>> nog.TraversalShortestPaths.go_for_vertices_in(None, None)
    Traceback (most recent call last):
    RuntimeError: Method go_for_vertices_in can only be called on a Traversal object.
    >>> nog.TraversalShortestPaths.go_for_distance_range(None, None, None)
    Traceback (most recent call last):
    RuntimeError: Method go_for_distance_range can only be called on a Traversal object.

    >>> nog.TraversalAStar.start_from(None, None)
    Traceback (most recent call last):
    RuntimeError: Method start_from can only be called on a Traversal object.
    >>> nog.TraversalAStar.__iter__(None)
    Traceback (most recent call last):
    RuntimeError: Method go can only be called on a Traversal object.
    >>> nog.TraversalAStar.go_to(None, None)
    Traceback (most recent call last):
    RuntimeError: Method go_to can only be called on a Traversal object.
    >>> nog.TraversalAStar.go_for_vertices_in(None, None)
    Traceback (most recent call last):
    RuntimeError: Method go_for_vertices_in can only be called on a Traversal object.

    >>> nog.TraversalMinimumSpanningTree.start_from(None)
    Traceback (most recent call last):
    RuntimeError: Method start_from can only be called on a Traversal object.
    >>> nog.TraversalMinimumSpanningTree.__iter__(None)
    Traceback (most recent call last):
    RuntimeError: Method go can only be called on a Traversal object.
    >>> nog.TraversalMinimumSpanningTree.go_to(None, None)
    Traceback (most recent call last):
    RuntimeError: Method go_to can only be called on a Traversal object.
    >>> nog.TraversalMinimumSpanningTree.go_for_vertices_in(None, None)
    Traceback (most recent call last):
    RuntimeError: Method go_for_vertices_in can only be called on a Traversal object.

    >>> nog.TraversalTopologicalSort.start_from(None)
    Traceback (most recent call last):
    RuntimeError: Method start_from can only be called on a Traversal object.
    >>> nog.TraversalTopologicalSort.__iter__(None)
    Traceback (most recent call last):
    RuntimeError: Method go can only be called on a Traversal object.
    >>> nog.TraversalTopologicalSort.go_to(None, None)
    Traceback (most recent call last):
    RuntimeError: Method go_to can only be called on a Traversal object.
    >>> nog.TraversalTopologicalSort.go_for_vertices_in(None, None)
    Traceback (most recent call last):
    RuntimeError: Method go_for_vertices_in can only be called on a Traversal object.
    """
    # inspection PyTypeChecker


class EdgeGadgets:
    # noinspection PyShadowingNames
    """
    # ------------ adapt_edge_index -------------
    >>> # Weighted directed edges given as mapping:
    >>> edges = {
    ...    'A': (('B', 1), ('C', 2)),
    ...    'B': (('C', 5), ('D', 5)),
    ...    'C': (('D', 2), ('E', 1), ('F', 5)),
    ...    'D': (('F', 2), ('G', 5)),
    ...    'E': (('F', 5), ('B', 1)),
    ...    'F': (('G', 2),)
    ... }
    >>> import nographs as nog
    >>> next_vertices = nog.adapt_edge_index(edges,add_inverted=False,labeled=True)
    >>> traversal = nog.TraversalShortestPaths(next_vertices)
    >>> vertex = traversal.start_from("A", build_paths=True).go_to("G")
    >>> print("vertex:", vertex, "distance:", traversal.distance,
    ...       "path edge count:", traversal.depth, "path:", traversal.paths[vertex])
    vertex: G distance: 8 path edge count: 4 path: ('A', 'C', 'D', 'F', 'G')
    >>>
    >>> # The generated next_vertices function returns the empty list if no edges where given
    >>> # for it in the dict.
    >>> next_vertices("G", None)
    ()

    >>> # Same edges, given in a sequence, node id is integer index of the sequence
    >>> edges = (
    ...     ((1, 1), (2, 2)),
    ...     ((2, 5), (3, 5)),
    ...     ((3, 2), (4, 1), (5, 5)),
    ...     ((5, 2), (6, 5)),
    ...     ((5, 5), (1, 1)),
    ...     ((6, 2),)
    ... )
    >>> next_vertices = nog.adapt_edge_index(edges,add_inverted=False,labeled=True)
    >>> traversal = nog.TraversalShortestPaths(next_vertices)
    >>> vertex = traversal.start_from(0, build_paths=True).go_to(6)
    >>> print("vertex:", vertex, "distance:", traversal.distance,
    ...       "path edge count:", traversal.depth, "path:", traversal.paths[vertex])
    vertex: 6 distance: 8 path edge count: 4 path: (0, 2, 3, 5, 6)
    >>>
    >>> # The generated next_vertices function returns the empty list if the index is out of range
    >>> next_vertices(7, None)
    ()

    >>> # Access testing data, where weighted undirected edges are given as dict
    >>> edges = {
    ...     'A': (('B', 1), ('C', 2)),
    ...     'B': (('E', 1), ('D', 4), ('F', 1)),
    ...     'C': (('D', 1),),
    ...     'D': (('E', 1),),
    ...     'E': (('F', 2),),
    ...     'F': (('G', 2),)
    ... }
    >>> next_vertices = nog.adapt_edge_index(edges,add_inverted=True,labeled=True)
    >>> traversal = nog.TraversalMinimumSpanningTree(next_vertices)
    >>> [traversal.edge for vertex in traversal.start_from("A")]
    [('A', 'B', 1), ('B', 'E', 1), ('B', 'F', 1), ('E', 'D', 1), ('D', 'C', 1), ('F', 'G', 2)]
    >>> # Note, that the minimum spanning tree contains edge ('D', 'C') in the direction not
    >>> # directly listed in the input.

    >>> # Access same testing graph, but the weighted undirected edges are given as list, vertices
    >>> # are int
    >>> edges = [
    ...     ((1, 1), (2, 2)),
    ...     ((4, 1), (3, 4), (5, 1)),
    ...     ((3, 1),),
    ...     ((4, 1),),
    ...     ((5, 2),),
    ...     ((6, 2),)
    ... ]
    >>> next_vertices = nog.adapt_edge_index(edges,add_inverted=True,labeled=True)
    >>> traversal = nog.TraversalMinimumSpanningTree(next_vertices)
    >>> [traversal.edge for vertex in traversal.start_from(0)]
    [(0, 1, 1), (1, 4, 1), (1, 5, 1), (4, 3, 1), (3, 2, 1), (5, 6, 2)]
    >>> # Note, that the minimum spanning tree contains edge (3, 2) in the direction not
    >>> # directly listed in the input.
    >>> next_vertices = nog.adapt_edge_index(set(),add_inverted=True,labeled=False)
    Traceback (most recent call last):
    ValueError: graph must be Mapping or Sequence

    # ------------ adapt_edge_iterable -------------
    >>> # Access testing data, where weighted directed edges are given as flat unsorted edge list
    >>> edges = [
    ...     ('A', 'B', 1), ('A', 'C', 2),
    ...     ('B', 'C', 5), ('B', 'D', 5),
    ...     ('C', 'D', 2), ('C', 'E', 1), ('C', 'F', 5),
    ...     ('D', 'F', 2), ('D', 'G', 5),
    ...     ('E', 'F', 5), ('E', 'B', 1),
    ...     ('F', 'G', 2)
    ... ]
    >>> import nographs as nog
    >>> next_vertices = nog.adapt_edge_iterable(edges,add_inverted=False,labeled=True)
    >>> traversal = nog.TraversalShortestPaths(next_vertices)
    >>> vertex = traversal.start_from("A", build_paths=True).go_to("G")
    >>> print("vertex:", vertex, "distance:", traversal.distance,
    ...       "path edge count:", traversal.depth, "path:", traversal.paths[vertex])
    vertex: G distance: 8 path edge count: 4 path: ('A', 'C', 'D', 'F', 'G')
    >>>
    >>> # Access testing data, where weighted undirected edges are given as flat edge list
    >>> edges = [
    ...     ('A', 'B', 1), ('A', 'C', 2),
    ...     ('B', 'E', 1), ('B', 'D', 4), ('B', 'F', 1),
    ...     ('C', 'D', 1),
    ...     ('D', 'E', 1),
    ...     ('E', 'F', 2),
    ...     ('F', 'G', 2)
    ... ]
    >>> next_vertices = nog.adapt_edge_iterable(edges,add_inverted=True,labeled=True)
    >>> traversal = nog.TraversalMinimumSpanningTree(next_vertices)
    >>> [traversal.edge for vertex in traversal.start_from("A")]
    [('A', 'B', 1), ('B', 'E', 1), ('B', 'F', 1), ('E', 'D', 1), ('D', 'C', 1), ('F', 'G', 2)]
    >>> # Note, that the minimum spanning tree contains edge ('D', 'C') in the direction not
    >>> # directly listed in the input.

    """
    # inspection PyShadowingNames


# --------- Tests for the gadgets --------
# (Only test cases that are not covered by examples in the documentation)

class ArrayTests:
    """
    Array positions should wrap at the position limits of each dimension:

    >>> array = nog.Array('''
    ... S..#.
    ... .#.#G
    ... #G...
    ... '''.strip().splitlines(), 2)
    >>> starts, goals = (array.findall(c) for c in "SG")
    >>> next_vertices = array.next_vertices_from_forbidden("#", wrap=True)
    >>> traversal = nog.TraversalBreadthFirst(next_vertices)
    >>> for found in traversal.start_from(start_vertices=starts, build_paths=True
    ...     ).go_for_vertices_in(goals):
    ...         traversal.depth, traversal.paths[found]
    (2, ((0, 0), (0, 4), (1, 4)))
    (2, ((0, 0), (0, 1), (2, 1)))

    "Diagonal" moves (edges) in the array positions should be allowed:
    >>> array = nog.Array('''
    ... S....
    ... ##...
    ... .G...
    ... '''.strip().splitlines(), 2)
    >>> start, goal = (array.findall(c)[0] for c in "SG")
    >>> next_vertices = array.next_vertices_from_forbidden("#", diagonals=True)
    >>> traversal = nog.TraversalBreadthFirst(next_vertices)
    >>> found = traversal.start_from(start, build_paths=True).go_to(goal)
    >>> traversal.depth, traversal.paths[found]
    (3, ((0, 0), (0, 1), (1, 2), (2, 1)))
    """

# if __name__ == "__main__":
#     import doctest
#
#     doctest.testmod()
